# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

# pylint: disable=unused-argument

from marshmallow import fields, post_load

from azure.ai.ml._restclient.v2023_04_01_preview.models import FeatureLags as FeatureLagsMode
from azure.ai.ml._restclient.v2023_04_01_preview.models import (
    ForecastHorizonMode,
    SeasonalityMode,
    ShortSeriesHandlingConfiguration,
    TargetAggregationFunction,
    TargetLagsMode,
    TargetRollingWindowSizeMode,
)
from azure.ai.ml._restclient.v2023_04_01_preview.models import UseStl as STLMode
from azure.ai.ml._schema.core.fields import StringTransformedEnum, UnionField
from azure.ai.ml._schema.core.schema import PatchedSchemaMeta


class ForecastingSettingsSchema(metaclass=PatchedSchemaMeta):
    country_or_region_for_holidays = fields.Str()
    cv_step_size = fields.Int()
    forecast_horizon = UnionField(
        [
            StringTransformedEnum(allowed_values=[ForecastHorizonMode.AUTO]),
            fields.Int(),
        ]
    )
    target_lags = UnionField(
        [
            StringTransformedEnum(allowed_values=[TargetLagsMode.AUTO]),
            fields.Int(),
            fields.List(fields.Int()),
        ]
    )
    target_rolling_window_size = UnionField(
        [
            StringTransformedEnum(allowed_values=[TargetRollingWindowSizeMode.AUTO]),
            fields.Int(),
        ]
    )
    time_column_name = fields.Str()
    time_series_id_column_names = UnionField([fields.Str(), fields.List(fields.Str())])
    frequency = fields.Str()
    feature_lags = StringTransformedEnum(allowed_values=[FeatureLagsMode.NONE, FeatureLagsMode.AUTO])
    seasonality = UnionField(
        [
            StringTransformedEnum(allowed_values=[SeasonalityMode.AUTO]),
            fields.Int(),
        ]
    )
    short_series_handling_config = StringTransformedEnum(
        allowed_values=[o.value for o in ShortSeriesHandlingConfiguration]
    )
    use_stl = StringTransformedEnum(allowed_values=[STLMode.NONE, STLMode.SEASON, STLMode.SEASON_TREND])
    target_aggregate_function = StringTransformedEnum(allowed_values=[o.value for o in TargetAggregationFunction])
    features_unknown_at_forecast_time = UnionField([fields.Str(), fields.List(fields.Str())])

    @post_load
    def make(self, data, **kwargs):
        from azure.ai.ml.entities._job.automl.tabular.forecasting_settings import ForecastingSettings

        return ForecastingSettings(**data)



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------
import copy
import enum
import json
import os
import shutil
import tempfile
from pathlib import Path
from typing import Dict
from zipfile import ZipFile

import pydash
import pytest
import yaml
from pytest_mock import MockFixture
from test_utilities.utils import build_temp_folder, mock_artifact_download_to_temp_directory, parse_local_path

from azure.ai.ml import load_component
from azure.ai.ml._internal._schema.component import NodeType
from azure.ai.ml._internal.entities.component import InternalComponent
from azure.ai.ml._internal.entities.spark import InternalSparkComponent
from azure.ai.ml._utils.utils import load_yaml
from azure.ai.ml.constants._common import AZUREML_INTERNAL_COMPONENTS_ENV_VAR
from azure.ai.ml.entities import Component
from azure.ai.ml.entities._builders.control_flow_node import LoopNode
from azure.ai.ml.exceptions import ValidationException

from .._utils import ANONYMOUS_COMPONENT_TEST_PARAMS, PARAMETERS_TO_TEST


class AdditionalIncludesCheckFunc(enum.Enum):
    """Enum for additional includes check function"""

    SKIP = 0
    SELF_IS_FILE = 1
    PARENT_EXISTS = 2
    NOT_EXISTS = 3
    NO_PARENT = 4


@pytest.mark.usefixtures("enable_internal_components")
@pytest.mark.unittest
@pytest.mark.pipeline_test
class TestComponent:
    def test_load_v2_component(self):
        yaml_path = "./tests/test_configs/components/helloworld_component.yml"
        load_component(yaml_path)

    def test_validate_internal_component(self):
        yaml_path = r"./tests/test_configs/internal/component_with_code/component_spec.yaml"
        from azure.ai.ml.entities._validate_funcs import validate_component

        validation_result = validate_component(yaml_path)
        assert validation_result.passed, repr(validation_result)

    def test_component_inputs_with_bool_and_date_value(self):
        yaml_path = (
            r"tests/test_configs/internal/command-component/command-linux/"
            r"component_with_bool_and_data_input/component.yaml"
        )
        component = load_component(yaml_path)
        assert component.inputs["bool_input"].default == "true"
        assert component.inputs["enum_input"].default == "true"
        assert component.inputs["enum_input"].enum == ["true", "false"]
        assert component.inputs["date_input"].default == "2023-01-01"

    def test_specific_error_message_on_load_from_dict(self):
        os.environ[AZUREML_INTERNAL_COMPONENTS_ENV_VAR] = "false"
        yaml_path = "./tests/test_configs/internal/helloworld/helloworld_component_command.yml"
        with pytest.raises(
            ValidationException,
            match="Internal components is a private feature in v2, " "please set environment variable",
        ):
            load_component(yaml_path, params_override=[{"type": "UnsupportedComponent"}])
        os.environ[AZUREML_INTERNAL_COMPONENTS_ENV_VAR] = "true"

    def test_load_from_registered_internal_command_component_rest_obj(self):
        base_component_internal = InternalComponent(
            _schema="https://componentsdk.azureedge.net/jsonschema/CommandComponent.json",
            name="ls_command",
            version="0.0.1",
            display_name="0.0.1",
            type=NodeType.COMMAND,
            tags={"category": "Component Tutorial", "contact": "amldesigner@microsoft.com"},
            is_deterministic=True,
            successful_return_code="Zero",
            inputs={"train_data": {"type": "path", "optional": False}},
            outputs={"output_dir": {"type": "path", "datastore_mode": "Upload"}},
            environment={"name": "AzureML-Minimal", "version": "45", "os": "Linux"},
            command="sh ls.sh {inputs.input_dir} {inputs.file_name} {outputs.output_dir}",
        )
        rest_obj = base_component_internal._to_rest_object()
        assert rest_obj.properties.component_spec == {
            "name": "ls_command",
            "tags": {"category": "Component Tutorial", "contact": "amldesigner@microsoft.com"},
            "type": "CommandComponent",
            "$schema": "https://componentsdk.azureedge.net/jsonschema/CommandComponent.json",
            "_source": "CLASS",
            "version": "0.0.1",
            "display_name": "0.0.1",
            "is_deterministic": True,
            "successful_return_code": "Zero",
            "inputs": {"train_data": {"type": "path", "optional": False}},
            "outputs": {"output_dir": {"type": "path", "datastore_mode": "Upload"}},
            "command": "sh ls.sh {inputs.input_dir} {inputs.file_name} {outputs.output_dir}",
            "environment": {"name": "AzureML-Minimal", "version": "45", "os": "Linux"},
        }
        component = Component._from_rest_object(rest_obj)
        assert component._to_dict() == {
            "name": "ls_command",
            "tags": {"category": "Component Tutorial", "contact": "amldesigner@microsoft.com"},
            "type": "CommandComponent",
            "$schema": "https://componentsdk.azureedge.net/jsonschema/CommandComponent.json",
            "version": "0.0.1",
            "display_name": "0.0.1",
            "is_deterministic": True,
            "successful_return_code": "Zero",
            "inputs": {"train_data": {"type": "path", "optional": False}},  # optional will be drop if False
            "outputs": {"output_dir": {"type": "path", "datastore_mode": "Upload"}},
            "command": "sh ls.sh {inputs.input_dir} {inputs.file_name} {outputs.output_dir}",
            "environment": {"name": "AzureML-Minimal", "version": "45", "os": "Linux"},
        }

    def test_load_from_registered_internal_scope_component_rest_obj(self):
        base_component_internal = InternalComponent(
            _schema="https://componentsdk.azureedge.net/jsonschema/CommandComponent.json",
            name="levance.convert2ss_31201029556679",
            version="85b54741.0bf9.4734.a5bb.0e469c7bf792",
            display_name="Convert Text to StructureStream",
            type=NodeType.SCOPE,
            tags={"org": "bing", "project": "relevance"},
            is_deterministic=True,
            inputs={
                "TextData": {
                    "type": "AnyFile",
                    "optional": False,
                    "description": "relative path on ADLS storage",
                },
                "ExtractionClause": {
                    "type": "string",
                    "optional": False,
                    "description": 'the extraction clause,something like "column1:string, column2:int"',
                },
            },
            outputs={"SSPath": {"type": "CosmosStructuredStream", "description": "output path  of ss"}},
            scope={
                "args": "RawData {inputs.TextData} SS_Data {outputs.SSPath} ExtractClause {inputs.ExtractionClause}",
                "script": "convert2ss.script",
            },
        )
        rest_obj = base_component_internal._to_rest_object()
        assert rest_obj.properties.component_spec == {
            "name": "levance.convert2ss_31201029556679",
            "tags": {"org": "bing", "project": "relevance"},
            "type": "ScopeComponent",
            "$schema": "https://componentsdk.azureedge.net/jsonschema/CommandComponent.json",
            "_source": "CLASS",
            "version": "85b54741.0bf9.4734.a5bb.0e469c7bf792",
            "display_name": "Convert Text to StructureStream",
            "is_deterministic": True,
            "inputs": {
                "TextData": {
                    "type": "AnyFile",
                    "optional": False,
                    "description": "relative path on ADLS storage",
                },
                "ExtractionClause": {
                    "type": "string",
                    "optional": False,
                    "description": 'the extraction clause,something like "column1:string, column2:int"',
                },
            },
            "outputs": {"SSPath": {"type": "CosmosStructuredStream", "description": "output path  of ss"}},
            "scope": {
                "args": "RawData {inputs.TextData} SS_Data {outputs.SSPath} ExtractClause {inputs.ExtractionClause}",
                "script": "convert2ss.script",
            },
        }
        component = Component._from_rest_object(rest_obj)
        assert component._to_dict() == {
            "name": "levance.convert2ss_31201029556679",
            "tags": {"org": "bing", "project": "relevance"},
            "type": "ScopeComponent",
            "$schema": "https://componentsdk.azureedge.net/jsonschema/CommandComponent.json",
            "version": "85b54741.0bf9.4734.a5bb.0e469c7bf792",
            "display_name": "Convert Text to StructureStream",
            "is_deterministic": True,
            "inputs": {
                "TextData": {
                    "type": "AnyFile",
                    "optional": False,
                    "description": "relative path on ADLS storage",
                },
                "ExtractionClause": {
                    "type": "string",
                    "optional": False,
                    "description": 'the extraction clause,something like "column1:string, column2:int"',
                },
            },
            "outputs": {"SSPath": {"type": "CosmosStructuredStream", "description": "output path  of ss"}},
            "scope": {
                "args": "RawData {inputs.TextData} SS_Data {outputs.SSPath} ExtractClause {inputs.ExtractionClause}",
                "script": "convert2ss.script",
            },
        }

    @pytest.mark.parametrize(
        "yaml_path,inputs,runsettings_dict,pipeline_runsettings_dict",
        PARAMETERS_TO_TEST,
    )
    def test_component_serialization(
        self,
        yaml_path: str,
        inputs: Dict,
        runsettings_dict: Dict,
        pipeline_runsettings_dict: Dict,
    ) -> None:
        with open(yaml_path, encoding="utf-8") as yaml_file:
            yaml_dict = yaml.safe_load(yaml_file)

        entity = load_component(yaml_path)

        expected_dict = copy.deepcopy(yaml_dict)

        # Linux is the default value of os in InternalEnvironment
        if "environment" in expected_dict:
            expected_dict["environment"]["os"] = "Linux"

        for input_port_name in expected_dict.get("inputs", {}):
            input_port = expected_dict["inputs"][input_port_name]
            # enum will be transformed to string
            if isinstance(input_port["type"], str) and input_port["type"].lower() in [
                "string",
                "enum",
                "float",
                "integer",
            ]:
                if "enum" in input_port:
                    input_port["enum"] = list(
                        map(lambda x: str(x).lower() if isinstance(x, bool) else str(x), input_port["enum"])
                    )
                if "default" in input_port:
                    input_port["default"] = (
                        str(input_port["default"]).lower()
                        if isinstance(input_port["default"], bool)
                        else str(input_port["default"])
                    )

        # code will be dumped as absolute path
        if "code" in expected_dict:
            expected_dict["code"] = parse_local_path(expected_dict["code"], entity.base_path)

        expected_dict["version"] = str(expected_dict["version"])

        if entity.type == "spark":
            expected_dict["jars"] = [expected_dict["jars"]]
            expected_dict["pyFiles"] = [expected_dict["pyFiles"]]
            expected_dict["environment"] = {
                "conda_file": {
                    "dependencies": ["python=3.8", {"pip": ["azureml-core==1.44.0", "shrike==1.31.2"]}],
                    "name": "component_env",
                },
                "image": "conda/miniconda3",
                "name": "CliV2AnonymousEnvironment",
                "version": entity.environment.version,
            }
        assert entity._to_dict() == expected_dict

        expected_rest_object = copy.deepcopy(expected_dict)
        expected_rest_object["_source"] = "YAML.COMPONENT"
        if entity.type == "spark":
            expected_rest_object["py_files"] = expected_rest_object.pop("pyFiles")
        rest_obj = entity._to_rest_object()
        assert rest_obj.properties.component_spec == expected_rest_object

        # inherit input type map from Component._from_rest_object
        for input_port in expected_dict.get("inputs", {}).values():
            if input_port["type"] == "String":
                input_port["type"] = input_port["type"].lower()

        try:
            from_rest_entity = InternalComponent._from_rest_object(rest_obj)
        except Exception as e:
            raise RuntimeError(
                "Failed to load component from rest object:\n{}\n"
                "Full error message:\n{}".format(json.dumps(rest_obj.as_dict(), indent=4), e)
            )
        assert from_rest_entity._to_dict() == expected_dict
        result = entity._validate()
        assert result._to_dict() == {"result": "Succeeded"}

    @pytest.mark.parametrize(
        "yaml_path,label",
        [
            ("preview_command_component.yaml", "1-preview"),
            ("legacy_distributed_component.yaml", "1-legacy"),
        ],
    )
    def test_command_mode_command_component(self, yaml_path: str, label: str):
        component = load_component("./tests/test_configs/internal/command-mode/{}".format(yaml_path))
        assert component._to_rest_object().properties.component_spec["type"] == f"{component.type}@{label}"

    def test_ipp_component_serialization(self):
        yaml_path = "./tests/test_configs/internal/ipp-component/spec.yaml"
        load_component(yaml_path)

    @pytest.mark.parametrize(
        "yaml_path,expected_dict",
        [
            (
                "./tests/test_configs/internal/env-conda-dependencies/component_spec.yaml",
                {
                    "conda": {
                        "conda_dependencies": {
                            "name": "project_environment",
                            "channels": ["defaults"],
                            "dependencies": [
                                "python=3.7.9",
                                "pip=20.0",
                                {"pip": ["azureml-defaults", "azureml-dataprep>=1.6"]},
                            ],
                        }
                    },
                    "os": "Linux",
                },
            ),
            (
                "./tests/test_configs/internal/env-pip-dependencies/component_spec.yaml",
                {
                    "conda": {
                        "conda_dependencies": {
                            "name": "project_environment",
                            "dependencies": [
                                "python=3.8.5",
                                {
                                    "pip": ["azureml-defaults", "rsa==4.7"],
                                },
                            ],
                        },
                    },
                    "os": "Linux",
                },
            ),
            (
                "./tests/test_configs/internal/env-dockerfile-build/component_spec.yaml",
                {
                    "docker": {
                        "build": {
                            "dockerfile": "FROM mcr.microsoft.com/azureml/openmpi4.1.0-ubuntu20.04:20220815.v1\n",
                        },
                    },
                    "python": {
                        "user_managed_dependencies": True,
                    },
                    "os": "Linux",
                },
            ),
        ],
    )
    def test_environment_dependencies_resolve(self, yaml_path: str, expected_dict: Dict) -> None:
        component: InternalComponent = load_component(source=yaml_path)
        component.environment.resolve(component._base_path)
        rest_obj = component._to_rest_object()
        assert rest_obj.properties.component_spec["environment"] == expected_dict
        assert component.environment._validate_docker_section(
            base_path=component.base_path, skip_path_validation=False
        ).passed

    @pytest.mark.parametrize(
        "yaml_path,invalid_field",
        [
            ("./tests/test_configs/internal/env-invalid/conda_file_not_exist.yaml", "conda_dependencies_file"),
            ("./tests/test_configs/internal/env-invalid/pip_file_not_exist.yaml", "pip_requirements_file"),
        ],
    )
    def test_invalid_environment(self, yaml_path: str, invalid_field: str) -> None:
        component = load_component(source=yaml_path)
        validation_result = component._validate()
        assert not validation_result.passed
        assert f"environment.conda.{invalid_field}" in validation_result.error_messages

    def test_environment_duplicate_dependencies_warning(self) -> None:
        yaml_path = "./tests/test_configs/internal/env-duplicate-dependencies/component_spec.yaml"
        component = load_component(source=yaml_path)
        validation_result = component._validate()
        # pass validate with warning message
        assert validation_result.passed
        expected_warning_message = (
            "environment.conda: Duplicated declaration of dependencies, "
            "will honor in the order conda_dependencies, conda_dependencies_file and pip_requirements_file."
        )
        assert str(validation_result._warnings[0]) == expected_warning_message

    def test_load_environment_with_version(self):
        yaml_path = (
            r"tests/test_configs/internal/command-component/command-linux/"
            r"component_with_bool_and_data_input/component.yaml"
        )
        yaml_dict = load_yaml(yaml_path)
        component = load_component(source=yaml_path)
        assert component.environment.name == yaml_dict["environment"]["name"]
        assert component.environment.version == str(yaml_dict["environment"]["version"])
        assert component.environment.os == yaml_dict["environment"]["os"]

    def test_resolve_local_code(self) -> None:
        # internal component code (snapshot) default includes items in base directory when code is None,
        # rather than COMPONENT_PLACEHOLDER in v2, this test targets to this case.
        from azure.ai.ml.entities._component.component import COMPONENT_PLACEHOLDER
        from azure.ai.ml.operations._component_operations import _try_resolve_code_for_component

        # create a mock function for param `get_arm_id_and_fill_back`
        def mock_get_arm_id_and_fill_back(asset, *args, **kwargs):
            return asset

        yaml_path = (
            "./tests/test_configs/internal/component_with_additional_includes/helloworld_additional_includes.yml"
        )
        component: InternalComponent = load_component(source=yaml_path)
        _try_resolve_code_for_component(
            component=component,
            resolver=mock_get_arm_id_and_fill_back,
        )
        assert component.code.path.name != COMPONENT_PLACEHOLDER

    def test_additional_includes(self) -> None:
        yaml_path = (
            "./tests/test_configs/internal/component_with_additional_includes/helloworld_additional_includes.yml"
        )
        component: InternalComponent = load_component(source=yaml_path)
        assert component._validate().passed, repr(component._validate())
        # resolve
        with component._build_code() as code:
            code_path: Path = code.path
            assert code_path.is_dir()
            assert (code_path / "LICENSE").is_file()
            assert (code_path / "library.zip").is_file()
            assert ZipFile(code_path / "library.zip").namelist() == ["library/", "library/hello.py", "library/world.py"]
            assert (code_path / "library1" / "hello.py").is_file()
            assert (code_path / "library1" / "world.py").is_file()
            assert code._ignore_file.is_file_excluded(code_path / "helloworld_additional_includes.additional_includes")

        assert not code_path.is_dir()

    @pytest.mark.parametrize(
        "test_files",
        [
            pytest.param(
                [
                    (
                        "component_with_additional_includes/.amlignore",
                        "test_ignore/*\nlibrary1/ignore.py",
                        AdditionalIncludesCheckFunc.SELF_IS_FILE,
                    ),
                    (
                        "component_with_additional_includes/test_ignore/a.py",
                        None,
                        AdditionalIncludesCheckFunc.NO_PARENT,
                    ),
                    # will be saved to library1/ignore.py, should be ignored
                    ("additional_includes/library1/ignore.py", None, AdditionalIncludesCheckFunc.NOT_EXISTS),
                    # will be saved to library1/ignore.py/a.py, should be ignored
                    # TODO: can't create with the same name?
                    # ("additional_includes/library1/ignore.py/a.py", None, AdditionalIncludesCheckFunc.NO_PARENT),
                    # will be saved to library1/test_ignore, should be kept
                    ("additional_includes/library1/test_ignore/a.py", None, AdditionalIncludesCheckFunc.SELF_IS_FILE),
                ],
                id="amlignore",
            ),
            pytest.param(
                [
                    # additional_includes for other spec, should be kept
                    (
                        "component_with_additional_includes/x.additional_includes",
                        None,
                        AdditionalIncludesCheckFunc.SELF_IS_FILE,
                    ),
                    (
                        "additional_includes/library1/x.additional_includes",
                        None,
                        AdditionalIncludesCheckFunc.SELF_IS_FILE,
                    ),
                    (
                        "additional_includes/library1/test/x.additional_includes",
                        None,
                        AdditionalIncludesCheckFunc.SELF_IS_FILE,
                    ),
                    # additional_includes in a different level, should be kept
                    (
                        "component_with_additional_includes/library2/helloworld_additional_includes.additional_includes",
                        None,
                        AdditionalIncludesCheckFunc.SELF_IS_FILE,
                    ),
                    (
                        "component_with_additional_includes/library2/library/helloworld_additional_includes.additional_includes",
                        None,
                        AdditionalIncludesCheckFunc.SELF_IS_FILE,
                    ),
                    # additional_includes in a different level in additional includes, should be kept
                    (
                        "additional_includes/library1/helloworld_additional_includes.additional_includes",
                        None,
                        AdditionalIncludesCheckFunc.SELF_IS_FILE,
                    ),
                ],
                id="additional_includes",
            ),
            pytest.param(
                [
                    ("component_with_additional_includes/hello.py", None, AdditionalIncludesCheckFunc.SELF_IS_FILE),
                    (
                        "component_with_additional_includes/test_code/.amlignore",
                        "hello.py",
                        AdditionalIncludesCheckFunc.SELF_IS_FILE,
                    ),
                    (
                        "component_with_additional_includes/test_code/hello.py",
                        None,
                        AdditionalIncludesCheckFunc.NOT_EXISTS,
                    ),
                    # shall we keep the empty folder?
                    (
                        "component_with_additional_includes/test_code/a/hello.py",
                        None,
                        AdditionalIncludesCheckFunc.NO_PARENT,
                    ),
                ],
                id="amlignore_subfolder",
            ),
            pytest.param(
                [
                    (
                        "additional_includes/library1/.amlignore",
                        "test_ignore\nignore.py",
                        AdditionalIncludesCheckFunc.SELF_IS_FILE,
                    ),
                    # will be saved to library1/ignore.py, should be ignored
                    ("additional_includes/library1/ignore.py", None, AdditionalIncludesCheckFunc.NOT_EXISTS),
                    # will be saved to library1/ignore.py/a.py, should be ignored
                    # TODO: can't create with the same name?
                    # ("additional_includes/library1/ignore.py/a.py", None, AdditionalIncludesCheckFunc.NO_PARENT),
                    # will be saved to library1/test_ignore, should be kept
                    ("additional_includes/library1/test_ignore/a.py", None, AdditionalIncludesCheckFunc.NOT_EXISTS),
                ],
                id="amlignore_in_additional_includes_folder",
            ),
            pytest.param(
                [
                    (
                        "additional_includes/library1/test_ignore/.amlignore",
                        "ignore.py",
                        AdditionalIncludesCheckFunc.SELF_IS_FILE,
                    ),
                    # will be saved to library1/ignore.py, should be ignored
                    (
                        "additional_includes/library1/test_ignore/ignore.py",
                        None,
                        AdditionalIncludesCheckFunc.NOT_EXISTS,
                    ),
                    (
                        "additional_includes/library1/test_ignore/ignore.py",
                        None,
                        AdditionalIncludesCheckFunc.NOT_EXISTS,
                    ),
                ],
                id="amlignore_in_additional_includes_subfolder",
            ),
            pytest.param(
                [
                    (
                        "component_with_additional_includes/__pycache__/a.pyc",
                        None,
                        AdditionalIncludesCheckFunc.NO_PARENT,
                    ),
                    (
                        "component_with_additional_includes/test/__pycache__/a.pyc",
                        None,
                        AdditionalIncludesCheckFunc.NO_PARENT,
                    ),
                    ("additional_includes/library1/__pycache__/a.pyc", None, AdditionalIncludesCheckFunc.NO_PARENT),
                    (
                        "additional_includes/library1/test/__pycache__/a.pyc",
                        None,
                        AdditionalIncludesCheckFunc.NO_PARENT,
                    ),
                ],
                id="pycache",
            ),
        ],
    )
    def test_additional_includes_advanced(self, test_files) -> None:
        with build_temp_folder(
            source_base_dir="./tests/test_configs/internal/",
            relative_dirs_to_copy=["component_with_additional_includes", "additional_includes"],
            extra_files_to_create={file: content for file, content, _ in test_files},
        ) as test_configs_dir:
            yaml_path = (
                Path(test_configs_dir) / "component_with_additional_includes" / "helloworld_additional_includes.yml"
            )

            component: InternalComponent = load_component(source=yaml_path)

            # resolve and check snapshot directory
            with component._build_code() as code:
                for file, content, check_func in test_files:
                    # original file is based on test_configs_dir, need to remove the leading
                    # "component_with_additional_includes" or "additional_includes" to get the relative path
                    resolved_file_path = Path(os.path.join(code.path, *Path(file).parts[1:]))
                    if check_func == AdditionalIncludesCheckFunc.NO_PARENT:
                        assert not resolved_file_path.parent.exists(), f"{file} should not have parent"
                    elif check_func == AdditionalIncludesCheckFunc.SELF_IS_FILE:
                        assert resolved_file_path.is_file(), f"{file} is not a file"
                        if content is not None:
                            assert resolved_file_path.read_text() == content, f"{file} content is not expected"
                    elif check_func == AdditionalIncludesCheckFunc.PARENT_EXISTS:
                        assert resolved_file_path.parent.is_dir(), f"{file} should have parent"
                    elif check_func == AdditionalIncludesCheckFunc.NOT_EXISTS:
                        assert not resolved_file_path.exists(), f"{file} should not exist"
                    elif check_func == AdditionalIncludesCheckFunc.SKIP:
                        pass
                    else:
                        raise ValueError(f"Unknown check func: {check_func}")

    def test_additional_includes_merge_folder(self) -> None:
        yaml_path = (
            "./tests/test_configs/internal/component_with_additional_includes/additional_includes_merge_folder.yml"
        )
        component: InternalComponent = load_component(source=yaml_path)
        assert component._validate().passed, repr(component._validate())
        with component._build_code() as code:
            code_path = code.path
            # first folder
            assert (code_path / "library1" / "__init__.py").is_file()
            assert (code_path / "library1" / "hello.py").is_file()
            # second folder content
            assert (code_path / "library1" / "utils").is_dir()
            assert (code_path / "library1" / "utils" / "__init__.py").is_file()
            assert (code_path / "library1" / "utils" / "salute.py").is_file()

    @pytest.mark.parametrize(
        "yaml_path,has_additional_includes",
        [
            ("no_code_and_additional_includes/component_spec.yml", False),
            ("code_only/component_spec.yml", False),
            ("code_and_additional_includes/component_spec.yml", True),
        ],
    )
    def test_additional_includes_with_code_specified(self, yaml_path: str, has_additional_includes: bool) -> None:
        yaml_path = os.path.join("./tests/test_configs/internal/component_with_additional_includes/", yaml_path)
        component: InternalComponent = load_component(source=yaml_path)
        assert component._validate().passed, repr(component._validate())
        # resolve
        with component._build_code() as code:
            code_path = code.path
            assert code_path.is_dir()
            if has_additional_includes:
                # additional includes is specified, code will be tmp folder and need to check each item
                # manually list here to avoid temp folder like __pycache__ breaking test.
                for path in [
                    "additional_includes_merge_folder.yml",
                    "code_and_additional_includes",
                    "code_only",
                    "helloworld_additional_includes.yml",
                    "helloworld_invalid_additional_includes_existing_file.yml",
                    "helloworld_invalid_additional_includes_root_directory.yml",
                    "helloworld_invalid_additional_includes_zip_file_not_found.yml",
                    "no_code_and_additional_includes",
                ]:
                    assert (code_path / path).is_file() if ".yml" in path else (code_path / path).is_dir()
                assert (code_path / "LICENSE").is_file()
            else:
                # additional includes not specified, code should be specified path (default yaml folder)
                yaml_dict = load_yaml(yaml_path)
                specified_code_path = Path(yaml_path).parent / yaml_dict.get("code", "./")
                assert code_path.resolve() == specified_code_path.resolve()

    def test_docker_file_in_additional_includes(self):
        yaml_path = (
            "./tests/test_configs/internal/component_with_dependency_" "in_additional_includes/with_docker_file.yml"
        )

        docker_file_path = "./tests/test_configs/internal/additional_includes/docker/DockerFile"
        with open(docker_file_path, "r") as docker_file:
            docker_file_content = docker_file.read()

        component: InternalComponent = load_component(source=yaml_path)
        assert component._validate().passed, repr(component._validate())
        with component._build_code():
            environment_rest_obj = component._to_rest_object().properties.component_spec["environment"]
            assert environment_rest_obj == {
                "docker": {
                    "build": {
                        "dockerfile": docker_file_content,
                    }
                },
                "os": "Linux",
            }

    def test_conda_pip_in_additional_includes(self):
        yaml_path = (
            "./tests/test_configs/internal/component_with_dependency_" "in_additional_includes/with_conda_pip.yml"
        )

        conda_file_path = "./tests/test_configs/internal/env-conda-dependencies/conda.yaml"
        with open(conda_file_path, "r") as conda_file:
            conda_file_content = yaml.safe_load(conda_file)

        component: InternalComponent = load_component(source=yaml_path)
        assert component._validate().passed, repr(component._validate())
        with component._build_code():
            environment_rest_obj = component._to_rest_object().properties.component_spec["environment"]
            assert environment_rest_obj == {
                "conda": {
                    "conda_dependencies": conda_file_content,
                },
                "os": "Linux",
            }

    def test_artifacts_in_additional_includes(self):
        with mock_artifact_download_to_temp_directory():
            yaml_path = "./tests/test_configs/internal/component_with_additional_includes/with_artifacts.yml"
            component: InternalComponent = load_component(source=yaml_path)
            assert component._validate().passed, repr(component._validate())
            with component._build_code() as code:
                code_path = code.path
                assert code_path.is_dir()
                for path in [
                    "version_1/",
                    "version_1/file",
                    "version_2/",
                    "version_2/file",
                    "file_version_1",
                    "file_version_2",
                    "DockerFile",
                ]:
                    assert (code_path / path).exists()

            yaml_path = (
                "./tests/test_configs/internal/component_with_additional_includes/"
                "artifacts_additional_includes_with_conflict.yml"
            )
            component: InternalComponent = load_component(source=yaml_path)
            with pytest.raises(
                RuntimeError,
                match="There are conflict files in additional include"
                ".*test_additional_include:version_1 in component-sdk-test-feed"
                ".*test_additional_include:version_3 in component-sdk-test-feed",
            ):
                with component._build_code():
                    pass

    @pytest.mark.parametrize(
        "yaml_path,expected_error_msg_prefix",
        [
            pytest.param(
                "helloworld_invalid_additional_includes_root_directory.yml",
                "Root directory is not supported for additional includes",
                id="root_as_additional_includes",
            ),
            pytest.param(
                "helloworld_invalid_additional_includes_existing_file.yml",
                "A file already exists for additional include",
                id="file_already_exists",
            ),
            pytest.param(
                "helloworld_invalid_additional_includes_zip_file_not_found.yml",
                "Unable to find additional include ../additional_includes/assets/LICENSE.zip",
                id="zip_file_not_found",
            ),
        ],
    )
    def test_invalid_additional_includes(self, yaml_path: str, expected_error_msg_prefix: str) -> None:
        component = load_component(
            os.path.join("./tests/test_configs/internal/component_with_additional_includes", yaml_path)
        )
        validation_result = component._validate()
        assert validation_result.passed is False
        assert validation_result.error_messages["*"].startswith(expected_error_msg_prefix)

    def test_component_input_types(self) -> None:
        yaml_path = "./tests/test_configs/internal/component_with_input_outputs/component_spec.yaml"
        component: InternalComponent = load_component(yaml_path)
        component.code = "scope:1"

        with open(yaml_path, "r") as f:
            yaml_dict = yaml.safe_load(f)
            for key, value in {
                "inputs.param_enum_cap.type": "enum",
            }.items():
                pydash.set_(yaml_dict, key, value)
        assert component._to_rest_object().properties.component_spec["inputs"] == yaml_dict["inputs"]
        assert component._to_rest_object().properties.component_spec["outputs"] == yaml_dict["outputs"]
        assert component._validate().passed is True, repr(component._validate())

        for key, value in {
            "inputs.param_bool_cap.type": "boolean",
            "inputs.param_int_cap.type": "integer",
            "inputs.param_string_cap.type": "string",
        }.items():
            pydash.set_(yaml_dict, key, value)
        regen_component = Component._from_rest_object(component._to_rest_object())
        assert regen_component._to_rest_object().properties.component_spec["inputs"] == yaml_dict["inputs"]
        assert regen_component._to_rest_object().properties.component_spec["outputs"] == yaml_dict["outputs"]
        assert component._validate().passed is True, repr(component._validate())

    def test_component_input_with_attrs(self) -> None:
        yaml_path = "./tests/test_configs/internal/component_with_input_outputs/component_spec_with_attrs.yaml"
        component: InternalComponent = load_component(source=yaml_path)

        expected_inputs = {
            "inputs": {
                "param_data_path": {
                    "description": "Path to the data",
                    "is_resource": True,
                    "datastore_mode": "mount",
                    "type": "path",
                },
                "param_bool": {"type": "boolean"},
                "param_enum_cap": {"enum": ["minimal", "reuse", "expiry", "policies"], "type": "enum"},
                "param_enum_with_int_values": {"default": "3", "enum": ["1", "2.0", "3", "4"], "type": "enum"},
                "param_float": {"type": "float", "default": "0.15", "max": "1.0", "min": "-1.0"},
                "param_int": {"type": "integer"},
                "param_string_with_default_value": {"default": ",", "type": "string"},
                "param_string_with_default_value_2": {"default": "utf8", "type": "string"},
                # yes will be converted to true in YAML 1.2, but we will load it as "yes" for backward compatibility
                "param_string_with_yes_value": {"default": "yes", "type": "string"},
                "param_string_with_quote_yes_value": {"default": "yes", "type": "string"},
            },
            "outputs": {
                "output_data_path": {
                    "datastore_mode": "mount",
                    "description": "Path to the data",
                    "is_link_mode": True,
                    "type": "path",
                }
            },
        }
        assert component._to_rest_object().properties.component_spec["inputs"] == expected_inputs["inputs"]
        assert component._to_rest_object().properties.component_spec["outputs"] == expected_inputs["outputs"]
        assert component._validate().passed is True, repr(component._validate())

        regenerated_component = Component._from_rest_object(component._to_rest_object())
        assert regenerated_component._to_rest_object().properties.component_spec["inputs"] == expected_inputs["inputs"]
        assert component._validate().passed is True, repr(component._validate())

    def test_component_output_with_attrs(self) -> None:
        yaml_path = "./tests/test_configs/internal/component_with_input_outputs/component_spec_with_outputs.yaml"
        component: InternalComponent = load_component(source=yaml_path)
        assert component

        expected_outputs = {
            "path_with_optional": {
                # unknown field optional will be ignored
                "type": "AnyDirectory",
            },
            "primitive_is_control": {"type": "boolean"},
        }
        assert component._to_rest_object().properties.component_spec["outputs"] == expected_outputs
        assert component._validate().passed is True, repr(component._validate())

        regenerated_component = Component._from_rest_object(component._to_rest_object())
        assert regenerated_component._to_rest_object().properties.component_spec["outputs"] == expected_outputs
        assert component._validate().passed is True, repr(component._validate())

    def test_component_input_list_type(self) -> None:
        yaml_path = "./tests/test_configs/internal/scope-component/component_spec.yaml"
        component: InternalComponent = load_component(yaml_path)
        assert component._validate().passed is True
        input_text_data_type = component._to_rest_object().properties.component_spec["inputs"]["TextData"]["type"]
        # for list type component input, REST object should remain type list for service contract
        assert isinstance(input_text_data_type, list)
        assert input_text_data_type == ["AnyFile", "AnyDirectory"]

    def test_loop_node_is_internal_components(self):
        from azure.ai.ml.constants._common import AZUREML_INTERNAL_COMPONENTS_ENV_VAR
        from azure.ai.ml.dsl._utils import environment_variable_overwrite

        yaml_path = "./tests/test_configs/internal/helloworld/helloworld_component_command.yml"
        component_func = load_component(source=yaml_path)
        loop_node = LoopNode(body=component_func())
        loop_node.body._referenced_control_flow_node_instance_id = loop_node._instance_id
        with environment_variable_overwrite(AZUREML_INTERNAL_COMPONENTS_ENV_VAR, "True"):
            validate_result = loop_node._validate_body()
            assert validate_result.passed

    @pytest.mark.parametrize(
        "relative_yaml_path,expected_snapshot_id",
        ANONYMOUS_COMPONENT_TEST_PARAMS,
    )
    def test_anonymous_component_reuse(self, relative_yaml_path: str, expected_snapshot_id: str):
        component: InternalComponent = load_component(
            source=Path("./tests/test_configs/internal/component-reuse/") / relative_yaml_path
        )
        with component._build_code() as code:
            assert code.name == expected_snapshot_id

            code.name = expected_snapshot_id
            with pytest.raises(
                AttributeError, match="InternalCode name are calculated based on its content and cannot be changed.*"
            ):
                code.name = expected_snapshot_id + "1"

    def test_snapshot_id_calculation(self):
        origin_test_configs_dir = Path("./tests/test_configs/internal/")
        with tempfile.TemporaryDirectory() as test_configs_dir:
            shutil.copytree(
                origin_test_configs_dir / "component-reuse/simple-command", Path(test_configs_dir) / "simple-command"
            )

            yaml_path = Path(test_configs_dir) / "simple-command" / "powershell_copy.yaml"

            component: InternalComponent = load_component(source=yaml_path)
            # create some files/folders expected to ignore
            code_pycache = yaml_path.parent / "__pycache__"
            code_pycache.mkdir()
            (code_pycache / "a.pyc").touch()

            # resolve and check snapshot directory
            with component._build_code() as code:
                # ANONYMOUS_COMPONENT_TEST_PARAMS[0] is the test params for simple-command
                assert code.name == ANONYMOUS_COMPONENT_TEST_PARAMS[0][1]

    def test_component_serialization_corner_case(self):
        yaml_path = "./tests/test_configs/internal/command-component-serialization-core-case/component_spec.yaml"
        component: InternalComponent = load_component(source=yaml_path)
        assert component
        rest_object = component._to_rest_object()
        assert rest_object.properties.component_spec == {
            "$schema": "https://componentsdk.azureedge.net/jsonschema/CommandComponent.json",
            "_source": "YAML.COMPONENT",
            "command": "echo {inputs.input_float} && echo {inputs.delimiter}",
            "display_name": "Hello Command",
            "environment": {"name": "AzureML-Designer", "os": "Linux"},
            "inputs": {
                "input_float": {"default": "1", "optional": False, "type": "Float"},  # previously this is 1.0
                "input_boolean": {"default": "False", "optional": False, "type": "Boolean"},
                "delimiter": {"default": "\t", "optional": True, "type": "String"},
            },
            "is_deterministic": True,
            "name": "hello_command",
            "type": "CommandComponent",
            "version": "0.10",  # previously this is 0.1
            "datatransfer": {"cloud_type": "aether"},
        }

    def test_load_from_internal_spark_component(self):
        yaml_path = PARAMETERS_TO_TEST[9].values[0]
        origin_component: InternalSparkComponent = load_component(source=yaml_path)
        base_rest_object = origin_component._to_rest_object()
        base_component: InternalSparkComponent = Component._from_rest_object(base_rest_object)
        expected_dict = base_component._to_dict()

        treat_rest_object = origin_component._to_rest_object()
        treat_rest_object.properties.component_spec["environment"] = {
            "os": "Linux",
            "conda": {
                "conda_dependencies": [
                    "python=3.8",
                    {
                        "pip": [
                            "azureml-core==1.44.0",
                            "shrike==1.31.2",
                        ],
                    },
                ],
            },
        }
        treat_component: InternalSparkComponent = Component._from_rest_object(treat_rest_object)

        for component in (base_component, treat_component, origin_component):
            assert isinstance(component, InternalSparkComponent)

        expected_dict["environment"]["version"] = treat_component.environment.version
        del expected_dict["environment"]["conda_file"]["name"]
        assert treat_component._to_dict() == expected_dict

    def test_load_from_internal_aether_bridge_component(self):
        yaml_path = "./tests/test_configs/internal/aether_bridge_component/component_spec.yaml"
        component: InternalComponent = load_component(source=yaml_path)
        assert component
        rest_object = component._to_rest_object()
        assert rest_object.properties.component_spec == {
            "name": "aether_bridge_component",
            "version": "0.0.1",
            "$schema": "https://componentsdk.azureedge.net/jsonschema/AetherBridgeComponent.json",
            "display_name": "Aether Bridge Component",
            "inputs": {
                "mock_param1": {"type": "AnyFile", "optional": False},
                "mock_param2": {"type": "AnyFile", "optional": False},
            },
            "aether": {"module_type": "ScrapingCloud", "ref_id": "mock_ref_id"},
            "outputs": {"job_info": {"type": "AnyFile"}},
            "type": "AetherBridgeComponent",
            "command": "mock.exe {inputs.mock_param1} {inputs.mock_param2} {outputs.job_info}",
            "_source": "YAML.COMPONENT",
        }

    def test_spark_component_special_type(self):
        yaml_path = "./tests/test_configs/internal/spark-component-special-type/spec.yaml"
        component: InternalSparkComponent = load_component(source=yaml_path)
        assert component
        rest_object = component._to_rest_object()
        assert rest_object.properties.component_spec["inputs"] == {
            "file_input1": {"datastore_mode": "hdfs", "description": "The data to be read.", "type": "path"},
            "file_input2": {"datastore_mode": "hdfs", "description": "The data to be read.", "type": "path"},
            "number_input": {
                "description": "The number to be used in the computation.",
                "max": "3.0",
                "min": "0.0",
                "type": "number",
            },
            "string_input": {"description": "The string to be used in the computation.", "type": "string"},
        }



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------
from typing import Optional


class ImageMetadata:
    """Metadata about the operating system image for the compute instance.

    :param is_latest_os_image_version: Specifies if the compute instance is running on the latest OS image version.
    :type is_latest_os_image_version: bool
    :param current_image_version: Version of the current image.
    :type current_image_version: str
    :param latest_image_version: The latest image version.
    :type latest_image_version: str

    .. admonition:: Example:

        .. literalinclude:: ../samples/ml_samples_compute.py
            :start-after: [START image_metadata]
            :end-before: [END image_metadata]
            :language: python
            :dedent: 8
            :caption: Creating a ImageMetadata object.
    """

    def __init__(
        self,
        *,
        is_latest_os_image_version: Optional[bool],
        current_image_version: Optional[str],
        latest_image_version: Optional[str]
    ) -> None:
        self._is_latest_os_image_version = is_latest_os_image_version
        self._current_image_version = current_image_version
        self._latest_image_version = latest_image_version

    @property
    def is_latest_os_image_version(self) -> Optional[bool]:
        """Whether or not a compute instance is running on the latest OS image version.

        :return: Boolean indicating if the compute instance is running the latest OS image version.
        :rtype: bool
        """
        return self._is_latest_os_image_version

    @property
    def current_image_version(self) -> Optional[str]:
        """The current OS image version number.

        :return: The current OS image version number.
        :rtype: str
        """
        return self._current_image_version

    @property
    def latest_image_version(self) -> Optional[str]:
        """The latest OS image version number.

        :return: The latest OS image version number.
        :rtype: str
        """
        return self._latest_image_version



# coding: utf-8

# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

"""
FILE: ml_samples_misc.py
DESCRIPTION:
    These samples demonstrate different ways to configure generic entities including jobs, assets, and components.
USAGE:
    python ml_samples_misc.py

"""

import os

from ml_samples_compute import handle_resource_exists_error

from azure.ai.ml import MLClient
from azure.identity import DefaultAzureCredential

subscription_id = os.environ["AZURE_SUBSCRIPTION_ID"]
resource_group = os.environ["RESOURCE_GROUP_NAME"]
workspace_name = "test-ws1"
credential = DefaultAzureCredential()
ml_client = MLClient(credential, subscription_id, resource_group, workspace_name=workspace_name)

import uuid

job_name = f"iris-dataset-job-{str(uuid.uuid4())}"


class MiscConfigurationOptions(object):
    def ml_misc_config_0(self):
        # [START job_operations_create_and_update]
        from azure.ai.ml import load_job

        created_job = ml_client.jobs.create_or_update(
            name=job_name,
            job=load_job(
                "./sdk/ml/azure-ai-ml/tests/test_configs/command_job/command_job_test_local_env.yml",
                params_override=[{"name": job_name}, {"compute": "cpucluster"}],
            ),
        )
        # [END job_operations_create_and_update]

        # [START job_operations_list]
        from azure.ai.ml._restclient.v2023_04_01_preview.models import ListViewType

        list_of_jobs = ml_client.jobs.list(parent_job_name=job_name, list_view_type=ListViewType.ARCHIVED_ONLY)
        # [END job_operations_list]

        # [START job_operations_get]
        retrieved_job = ml_client.jobs.get(job_name)
        # [END job_operations_get]

        # [START job_operations_begin_cancel]
        cancel_poller = ml_client.jobs.begin_cancel(job_name)
        print(cancel_poller.result())
        # [END job_operations_begin_cancel]

        # [START job_operations_validate]
        from azure.ai.ml import load_job
        from azure.ai.ml.entities import PipelineJob

        pipeline_job: PipelineJob = load_job(  # type:ignore
            "./sdk/ml/azure-ai-ml/tests/test_configs/pipeline_jobs/invalid/combo.yml",
            params_override=[{"name": job_name}, {"compute": "cpucluster"}],
        )
        print(ml_client.jobs.validate(pipeline_job).error_messages)
        # [END job_operations_validate]

        # [START job_operations_archive]
        ml_client.jobs.archive(name=job_name)
        # [END job_operations_archive]

        # [START job_operations_restore]
        ml_client.jobs.restore(name=job_name)
        # [END job_operations_restore]

        # [START job_operations_stream_logs]
        running_job = ml_client.jobs.create_or_update(
            load_job(
                "./sdk/ml/azure-ai-ml/tests/test_configs/command_job/command_job_test_local_env.yml",
                params_override=[{"name": job_name}, {"compute": "cpucluster"}],
            )
        )
        ml_client.jobs.stream(running_job.name)
        # [END job_operations_stream_logs]

        # [START job_operations_download]
        ml_client.jobs.download(name=job_name, download_path="./job-1-logs", all=True)
        # [END job_operations_download]

        # [START model_entity_create]
        from azure.ai.ml.entities import Model

        model = Model(
            name="model1",
            version="5",
            description="my first model in prod",
            path="models/very_important_model.pkl",
            properties={"prop1": "value1", "prop2": "value2"},
            type="mlflow_model",
            flavors={
                "sklearn": {"sklearn_version": "0.23.2"},
                "python_function": {"loader_module": "office.plrmodel", "python_version": 3.6},
            },
            stage="Production",
        )
        ml_client.models.create_or_update(model)
        # [END model_entity_create]

        # [START model_operations_archive]
        ml_client.models.archive(name="model1", version="5")
        # [END model_operations_archive]

        # [START model_operations_restore]
        ml_client.models.restore(name="model1", version="5")
        # [END model_operations_restore]

        # [START model_batch_deployment_settings_entity_create]
        from azure.ai.ml.entities._deployment.model_batch_deployment_settings import ModelBatchDeploymentSettings

        modelBatchDeploymentSetting = ModelBatchDeploymentSettings(
            mini_batch_size=256,
            instance_count=5,
            max_concurrency_per_instance=2,
            output_file_name="output-file-name",
            environment_variables={"env1": "value1", "env2": "value2"},
            error_threshold=2,
            logging_level="1",
        )
        # [END model_batch_deployment_settings_entity_create]

        # [START model_configuration_entity_create]
        from azure.ai.ml.entities._assets._artifacts._package.model_configuration import ModelConfiguration

        modelConfiguration = ModelConfiguration(mode="model-mode", mount_path="model-mount-path")
        # [END model_configuration_entity_create]

        # [START model_package_input_entity_create]
        from azure.ai.ml.entities._assets._artifacts._package.model_package import ModelPackageInput

        modelPackageInput = ModelPackageInput(type="input-type", mode="input-mode", mount_path="input-mount-path")
        # [END model_package_input_entity_create]

        # [START model_package_entity_create]
        from azure.ai.ml.entities import AzureMLOnlineInferencingServer, CodeConfiguration, ModelPackage

        modelPackage = ModelPackage(  # type:ignore
            inferencing_server=AzureMLOnlineInferencingServer(
                code_configuration=CodeConfiguration(code="../model-1/foo/", scoring_script="score.py")
            ),
            target_environment_name="env-name",
            target_environment_version="1.0",
            environment_variables={"env1": "value1", "env2": "value2"},
            tags={"tag1": "value1", "tag2": "value2"},
        )
        # [END model_package_entity_create]

        # [START create_inputs_outputs]
        from azure.ai.ml import Input, Output
        from azure.ai.ml.entities import CommandJob, CommandJobLimits

        command_job = CommandJob(
            code="./src",
            command="python train.py --ss {search_space.ss}",
            inputs={
                "input1": Input(path="trial.csv", mode="ro_mount", description="trial input data"),
                "input_2": Input(
                    path="azureml:list_data_v2_test:2", type="uri_folder", description="registered data asset"
                ),
            },
            outputs={"default": Output(path="./foo")},
            compute="trial",
            environment="AzureML-sklearn-1.0-ubuntu20.04-py38-cpu:33",
            limits=CommandJobLimits(timeout=120),
        )
        # [END create_inputs_outputs]

        # [START load_job]
        from azure.ai.ml import load_job

        job = load_job(source="./sdk/ml/azure-ai-ml/tests/test_configs/command_job/command_job_test_local_env.yml")
        # [END load_job]

        # [START load_model]
        from azure.ai.ml import load_model

        model = load_model(
            source="./sdk/ml/azure-ai-ml/tests/test_configs/model/model_with_stage.yml",
            params_override=[{"name": "new_model_name"}, {"version": "1"}],
        )
        # [END load_model]

        # [START load_model_package]
        from azure.ai.ml import load_model_package

        model_package = load_model_package(
            "./sdk/ml/azure-ai-ml/tests/test_configs/model_package/model_package_simple.yml"
        )
        # [END load_model_package]

        # [START tensorflow_distribution_configuration]
        from azure.ai.ml import TensorFlowDistribution
        from azure.ai.ml.entities import CommandComponent

        component = CommandComponent(
            name="microsoftsamples_tf",
            description="This is the TF command component",
            inputs={
                "component_in_number": {"description": "A number", "type": "number", "default": 10.99},
                "component_in_path": {"description": "A path", "type": "uri_folder"},
            },
            outputs={"component_out_path": {"type": "uri_folder"}},
            command="echo Hello World & echo ${{inputs.component_in_number}} & echo ${{inputs.component_in_path}} "
            "& echo ${{outputs.component_out_path}}",
            environment="AzureML-sklearn-1.0-ubuntu20.04-py38-cpu:33",
            distribution=TensorFlowDistribution(
                parameter_server_count=1,
                worker_count=2,
            ),
            instance_count=2,
        )
        # [END tensorflow_distribution_configuration]

        # [START pytorch_distribution_configuration]
        from azure.ai.ml import PyTorchDistribution
        from azure.ai.ml.entities import CommandComponent

        component = CommandComponent(
            name="microsoftsamples_torch",
            description="This is the PyTorch command component",
            inputs={
                "component_in_number": {"description": "A number", "type": "number", "default": 10.99},
                "component_in_path": {"description": "A path", "type": "uri_folder"},
            },
            outputs={"component_out_path": {"type": "uri_folder"}},
            command="echo Hello World & echo ${{inputs.component_in_number}} & echo ${{inputs.component_in_path}} "
            "& echo ${{outputs.component_out_path}}",
            environment="AzureML-sklearn-1.0-ubuntu20.04-py38-cpu:33",
            distribution=PyTorchDistribution(
                process_count_per_instance=2,
            ),
            instance_count=2,
        )
        # [END pytorch_distribution_configuration]

        # [START mpi_distribution_configuration]
        from azure.ai.ml import MpiDistribution
        from azure.ai.ml.entities import CommandComponent

        component = CommandComponent(
            name="microsoftsamples_mpi",
            description="This is the MPI command component",
            inputs={
                "component_in_number": {"description": "A number", "type": "number", "default": 10.99},
                "component_in_path": {"description": "A path", "type": "uri_folder"},
            },
            outputs={"component_out_path": {"type": "uri_folder"}},
            command="echo Hello World & echo ${{inputs.component_in_number}} & echo ${{inputs.component_in_path}} "
            "& echo ${{outputs.component_out_path}}",
            environment="AzureML-sklearn-1.0-ubuntu20.04-py38-cpu:33",
            distribution=MpiDistribution(
                process_count_per_instance=2,
            ),
            instance_count=2,
        )
        # [END mpi_distribution_configuration]

        # [START code_configuration]
        from azure.ai.ml.entities import BatchDeployment, CodeConfiguration

        deployment = BatchDeployment(
            name="non-mlflow-deployment",
            description="this is a sample non-mlflow deployment",
            endpoint_name="my-batch-endpoint",
            model=model,
            code_configuration=CodeConfiguration(
                code="configs/deployments/model-2/onlinescoring", scoring_script="score1.py"
            ),
            environment="env",
            compute="cpu-cluster",
            instance_count=2,
            max_concurrency_per_instance=2,
            mini_batch_size=10,
            output_file_name="predictions.csv",
        )
        # [END code_configuration]

        # [START intellectual_property_configuration]
        from azure.ai.ml.constants import IPProtectionLevel
        from azure.ai.ml.entities import CommandComponent, IntellectualProperty

        component = CommandComponent(
            name="random_name",
            version="1",
            environment="azureml:AzureML-Minimal:1",
            command="echo hello",
            intellectual_property=IntellectualProperty(publisher="contoso", protection_level=IPProtectionLevel.ALL),
        )
        # [END intellectual_property_configuration]

        # [START personal_access_token_configuration]
        from azure.ai.ml.entities import PatTokenConfiguration, WorkspaceConnection

        ws_connection = WorkspaceConnection(
            target="my_target",
            type="python_feed",
            credentials=PatTokenConfiguration(pat="abcdefghijklmnopqrstuvwxyz"),
            name="my_connection",
            metadata=None,
        )
        # [END personal_access_token_configuration]

        # [START job_schedule_configuration]
        from azure.ai.ml import load_job
        from azure.ai.ml.entities import JobSchedule, RecurrencePattern, RecurrenceTrigger

        pipeline_job = load_job(
            "./sdk/ml/azure-ai-ml/tests/test_configs/command_job/command_job_test_local_env.yml"
        )  # type:ignore
        trigger = RecurrenceTrigger(
            frequency="week",
            interval=4,
            schedule=RecurrencePattern(hours=10, minutes=15, week_days=["Monday", "Tuesday"]),
            start_time="2023-03-10",
        )
        job_schedule = JobSchedule(name="simple_sdk_create_schedule", trigger=trigger, create_job=pipeline_job)
        # [END job_schedule_configuration]

        # [START cron_trigger_configuration]
        from datetime import datetime

        from azure.ai.ml.constants import TimeZone
        from azure.ai.ml.entities import CronTrigger

        trigger = CronTrigger(  # type:ignore
            expression="15 10 * * 1",
            start_time=datetime(year=2022, month=3, day=10, hour=10, minute=15),
            end_time=datetime(year=2022, month=6, day=10, hour=10, minute=15),
            time_zone=TimeZone.PACIFIC_STANDARD_TIME,
        )
        # [END cron_trigger_configuration]

        # [START resource_requirements_configuration]
        from azure.ai.ml.entities import (
            CodeConfiguration,
            KubernetesOnlineDeployment,
            ResourceRequirementsSettings,
            ResourceSettings,
        )

        blue_deployment = KubernetesOnlineDeployment(
            name="kubernetes_deployment",
            endpoint_name="online_endpoint_name",
            model=load_model("./sdk/ml/azure-ai-ml/tests/test_configs/model/model_with_stage.yml"),
            environment="azureml:AzureML-Minimal:1",
            code_configuration=CodeConfiguration(
                code="endpoints/online/model-1/onlinescoring", scoring_script="score.py"
            ),
            instance_count=1,
            resources=ResourceRequirementsSettings(
                requests=ResourceSettings(
                    cpu="500m",
                    memory="0.5Gi",
                ),
                limits=ResourceSettings(
                    cpu="1",
                    memory="1Gi",
                ),
            ),
        )
        # [END resource_requirements_configuration]

        # [START ssh_job_service_configuration]
        from azure.ai.ml import command
        from azure.ai.ml.entities import JupyterLabJobService, SshJobService, TensorBoardJobService, VsCodeJobService

        node = command(
            name="interactive-command-job",
            description="description",
            environment="AzureML-sklearn-1.0-ubuntu20.04-py38-cpu:33",
            command="ls",
            compute="testCompute",
            services={
                "my_ssh": SshJobService(),
                "my_tensorboard": TensorBoardJobService(log_dir="~/blog"),
                "my_jupyter_lab": JupyterLabJobService(),
                "my_vscode": VsCodeJobService(),
            },
        )
        # [END ssh_job_service_configuration]

        # [START build_context_entity_create]
        from azure.ai.ml.entities._assets.environment import BuildContext

        build_context = BuildContext(dockerfile_path="docker-file-path", path="docker-build-context-path")
        # [END build_context_entity_create]

        # [START base_env_entity_create]
        from azure.ai.ml.entities._assets._artifacts._package.base_environment_source import BaseEnvironment

        base_environment = BaseEnvironment(type="base-env-type", resource_id="base-env-resource-id")
        # [END base_env_entity_create]

        # [START env_entity_create]
        from azure.ai.ml.entities._assets.environment import Environment

        environment = Environment(
            name="env-name",
            version="2.0",
            description="env-description",
            image="env-image",
            conda_file="./sdk/ml/azure-ai-ml/tests/test_configs/deployments/model-1/environment/conda.yml",
            tags={"tag1": "value1", "tag2": "value2"},
            properties={"prop1": "value1", "prop2": "value2"},
            datastore="datastore",
        )
        # [END env_entity_create]

        # [START env_operations_create_or_update]
        from azure.ai.ml.entities import BuildContext, Environment

        env_docker_context = Environment(
            build=BuildContext(
                path="./sdk/ml/azure-ai-ml/tests/test_configs/environment/environment_files",
                dockerfile_path="DockerfileNonDefault",
            ),
            name="create-environment",
            version="2.0",
            description="Environment created from a Docker context.",
        )
        ml_client.environments.create_or_update(env_docker_context)
        # [END env_operations_create_or_update]

        # [START env_entities_validate]
        from azure.ai.ml.entities import BuildContext, Environment

        env_docker_context = Environment(
            build=BuildContext(
                path="./sdk/ml/azure-ai-ml/tests/test_configs/environment/environment_files",
                dockerfile_path="DockerfileNonDefault",
            ),
            name="create-environment",
            version="2.0",
            description="Environment created from a Docker context.",
        )

        env_docker_context.validate()
        # [END env_entities_validate]

        # [START env_operations_archive]
        ml_client.environments.archive("create-environment", "2.0")
        # [END env_operations_archive]

        # [START env_operations_restore]
        ml_client.environments.restore("create-environment", "2.0")
        # [END env_operations_restore]

        # [START env_operations_list]
        ml_client.environments.list()
        # [END env_operations_list]

        # [START env_operations_get]
        ml_client.environments.get("create-environment", "2.0")
        # [END env_operations_get]

    @handle_resource_exists_error
    def ml_misc_config_1(self):
        from random import randint

        from azure.ai.ml import load_batch_endpoint
        from azure.ai.ml.entities import BatchEndpoint

        endpoint_example = load_batch_endpoint(
            source="./sdk/ml/azure-ai-ml/tests/test_configs/endpoints/batch/batch_endpoint_mlflow_new.yaml",
            params_override=[{"name": f"endpoint-{randint(0, 1000)}"}],
        )
        ml_client.batch_endpoints.begin_create_or_update(endpoint_example)
        endpoint_name = endpoint_example.name

        # [START batch_deployment_operations_begin_create_or_update]
        from azure.ai.ml import load_batch_deployment
        from azure.ai.ml.entities import BatchDeployment

        deployment_example = load_batch_deployment(
            source="./sdk/ml/azure-ai-ml/tests/test_configs/deployments/batch/batch_deployment_anon_env_with_image.yaml",
            params_override=[{"name": f"deployment-{randint(0, 1000)}", "endpoint_name": endpoint_example.name}],
        )

        ml_client.batch_deployments.begin_create_or_update(deployment=deployment_example, skip_script_validation=True)
        # [END batch_deployment_operations_begin_create_or_update]

        deployment_name = deployment_example.name

        # [START batch_deployment_operations_get]
        ml_client.batch_deployments.get(deployment_name, endpoint_name)
        # [END batch_deployment_operations_get]

        # [START batch_deployment_operations_list]
        ml_client.batch_deployments.list(endpoint_name)
        # [END batch_deployment_operations_list]

        # [START batch_deployment_operations_list_jobs]
        ml_client.batch_deployments.list_jobs(deployment_name, endpoint_name)
        # [END batch_deployment_operations_list_jobs]

        # [START batch_deployment_operations_delete]
        ml_client.batch_deployments.begin_delete(deployment_name, endpoint_name)
        # [END batch_deployment_operations_delete]

        # [START batch_endpoint_operations_list]
        ml_client.batch_endpoints.list()
        # [END batch_endpoint_operations_list]

        # [START batch_endpoint_operations_get]
        ml_client.batch_endpoints.get(endpoint_name)
        # [END batch_endpoint_operations_get]

        # [START batch_endpoint_operations_delete]
        ml_client.batch_endpoints.begin_delete(endpoint_name)
        # [END batch_endpoint_operations_delete]

        from random import randint

        endpoint_name_2 = f"new-endpoint-{randint(0, 1000)}"

        # [START batch_endpoint_operations_create_or_update]
        from azure.ai.ml.entities import BatchEndpoint

        endpoint_example = BatchEndpoint(name=endpoint_name_2)
        ml_client.batch_endpoints.begin_create_or_update(endpoint_example)
        # [END batch_endpoint_operations_create_or_update]

        # [START batch_endpoint_operations_invoke]
        ml_client.batch_endpoints.invoke(endpoint_name_2)
        # [END batch_endpoint_operations_invoke]

        # [START batch_endpoint_operations_list_jobs]
        ml_client.batch_endpoints.list_jobs(endpoint_name_2)
        # [END batch_endpoint_operations_list_jobs]

    def ml_misc_config_2(self):
        # [START component_operations_create_or_update]
        from azure.ai.ml import load_component
        from azure.ai.ml.entities._component.component import Component

        component_example = load_component(
            source="./sdk/ml/azure-ai-ml/tests/test_configs/components/helloworld_component.yml",
            params_override=[{"version": "1.0.2"}],
        )
        component = ml_client.components.create_or_update(component_example)
        # [END component_operations_create_or_update]
        print(component)

        # [START code_operations_create_or_update]
        from azure.ai.ml.entities._assets._artifacts.code import Code

        code_example = Code(name="my-code-asset", version="2.0", path="./sdk/ml/azure-ai-ml/samples/src")
        code_asset = ml_client._code.create_or_update(code_example)
        # [END code_operations_create_or_update]

        from random import randint

        data_asset_name = f"data_asset_name_{randint(0, 1000)}"
        # [START data_operations_create_or_update]
        from azure.ai.ml.entities import Data

        data_asset_example = Data(name=data_asset_name, version="2.0", path="./sdk/ml/azure-ai-ml/samples/src")
        ml_client.data.create_or_update(data_asset_example)
        # [END data_operations_create_or_update]

        # [START component_operations_list]
        print(ml_client.components.list())
        # [END component_operations_list]

        # [START component_operations_get]
        ml_client.components.get(name=component_example.name, version="1.0.2")
        # [END component_operations_get]

        # [START component_operations_validate]
        from azure.ai.ml.entities._component.component import Component

        ml_client.components.validate(component_example)
        # [END component_operations_validate]

        # [START component_operations_archive]
        ml_client.components.archive(name=component_example.name)
        # [END component_operations_archive]

        # [START component_operations_restore]
        ml_client.components.restore(name=component_example.name)
        # [END component_operations_restore]

        # [START code_operations_get]
        ml_client._code.get(name=code_asset.name, version=code_asset.version)
        # [END code_operations_get]

        # [START data_operations_list]
        ml_client.data.list(name="data_asset_name")
        # [END data_operations_list]

        # [START data_operations_get]
        ml_client.data.get(name="data_asset_name", version="2.0")
        # [END data_operations_get]

        # [START data_operations_import_data]
        from azure.ai.ml.entities._data_import.data_import import DataImport
        from azure.ai.ml.entities._inputs_outputs.external_data import Database

        database_example = Database(query="SELECT ID FROM DataTable", connection="azureml:my_azuresqldb_connection")
        data_import_example = DataImport(
            name="data_asset_name", path="azureml://datastores/workspaceblobstore/paths/", source=database_example
        )
        ml_client.data.import_data(data_import_example)
        # [END data_operations_import_data]

        # [START data_operations_list_materialization_status]
        ml_client.data.list_materialization_status("data_asset_name")
        # [END data_operations_list_materialization_status]

        # [START data_operations_archive]
        ml_client.data.archive("data_asset_name")
        # [END data_operations_archive]

        # [START data_operations_restore]
        ml_client.data.restore("data_asset_name")
        # [END data_operations_restore]

        try:
            # [START data_operations_share]
            ml_client.data.share(
                name="data_asset_name",
                version="2.0",
                registry_name="my-registry",
                share_with_name="transformed-nyc-taxi-data-shared-from-ws",
                share_with_version="2.0",
            )
        # [END data_operations_share]
        except TypeError:
            pass

        # [START datastore_operations_create_or_update]
        from azure.ai.ml.entities import AzureBlobDatastore

        datastore_example = AzureBlobDatastore(
            name="azure_blob_datastore",
            account_name="sdkvnextclidcdnrc7zb7xyy",  # cspell:disable-line
            container_name="testblob",
        )
        ml_client.datastores.create_or_update(datastore_example)
        # [END datastore_operations_create_or_update]

        # [START datastore_operations_list]
        ml_client.datastores.list()
        # [END datastore_operations_list]

        # [START datastore_operations_get]
        ml_client.datastores.get("azure_blob_datastore")
        # [END datastore_operations_get]

        # [START datastore_operations_get_default]
        ml_client.datastores.get_default()
        # [END datastore_operations_get_default]

        # [START datastore_operations_delete]
        ml_client.datastores.delete("azure_blob_datastore")
        # [END datastore_operations_delete]

        # [START validation_result]
        """For example, if repr(self) is:
        ```python
            {
                "errors": [
                    {
                        "path": "jobs.job_a.inputs.input_str",
                        "message": "input_str is required",
                        "value": None,
                    },
                    {
                        "path": "jobs.job_a.inputs.input_str",
                        "message": "input_str must be in the format of xxx",
                        "value": None,
                    },
                    {
                        "path": "settings.on_init",
                        "message": "On_init job name job_b does not exist in jobs.",
                        "value": None,
                    },
                ],
                "warnings": [
                    {
                        "path": "jobs.job_a.inputs.input_str",
                        "message": "input_str is required",
                        "value": None,
                    }
                ]
            }
            ```
            then the error_messages will be:
            ```python
            {
                "jobs.job_a.inputs.input_str": "input_str is required; input_str must be in the format of xxx",
                "settings.on_init": "On_init job name job_b does not exist in jobs.",
            }
            ```
            """
        # [END validation_result]

    @handle_resource_exists_error
    def ml_misc_config_3(self):
        # [START job_operations_show_services]
        job_services = ml_client.jobs.show_services(job_name)
        # [END job_operations_show_services]


if __name__ == "__main__":
    sample = MiscConfigurationOptions()
    sample.ml_misc_config_0()
    sample.ml_misc_config_1()
    sample.ml_misc_config_2()
    sample.ml_misc_config_3()



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------
"""This module simulate run() which can specify succeed every n items from argument."""
import argparse
from pathlib import Path


def init():
    """Init."""
    global OUTPUT_PATH

    parser = argparse.ArgumentParser(allow_abbrev=False, description="ParallelRunStep Agent")
    parser.add_argument("--job_output_path", type=str, default=0)
    args, _ = parser.parse_known_args()
    OUTPUT_PATH = args.job_output_path
    print("Pass through init done")


def run(mini_batch):
    """Run."""

    for file_path in mini_batch:
        file = Path(file_path)
        print("Processing {}".format(file))
        assert file.exists()

        # Two customers reported transient error when using OutputFileDatasetConfig.
        # It hits "FileNotFoundError" when writing to a file in the output_dir folder,
        #  even the folder did exist per logs.
        # This is to simulate such case and hope we can repro in our gated build.
        output_dir = Path(OUTPUT_PATH)
        print("output_dir", output_dir)
        print("output_dir exits", Path(output_dir).exists())
        (Path(output_dir) / file.name).write_text(file_path)

    return mini_batch



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

import os
from os import PathLike
from pathlib import Path
from typing import IO, Any, AnyStr, Dict, List, Optional, Union

from azure.ai.ml._restclient.v2025_01_01_preview.models._models_py3 import CapabilityHost as RestCapabilityHost
from azure.ai.ml._restclient.v2025_01_01_preview.models._models_py3 import (
    CapabilityHostProperties as RestCapabilityHostProperties,
)
from azure.ai.ml._schema.workspace.ai_workspaces.capability_host import CapabilityHostSchema
from azure.ai.ml._utils._experimental import experimental
from azure.ai.ml._utils.utils import dump_yaml_to_file
from azure.ai.ml.constants._common import BASE_PATH_CONTEXT_KEY, PARAMS_OVERRIDE_KEY
from azure.ai.ml.constants._workspace import CapabilityHostKind
from azure.ai.ml.entities._resource import Resource
from azure.ai.ml.entities._util import load_from_dict


@experimental
class CapabilityHost(Resource):
    """Initialize a CapabilityHost instance.
    Capabilityhost management is controlled by MLClient's capabilityhosts operations.

    :param name: The name of the capability host.
    :type name: str
    :param description: The description of the capability host.
    :type description: Optional[str]
    :param vector_store_connections: A list of vector store  (AI Search) connections.
    :type vector_store_connections: Optional[List[str]]
    :param ai_services_connections: A list of OpenAI service connection.
    :type ai_services_connections: Optional[List[str]]
    :param storage_connections: A list of storage connections. Default storage connection value is
        projectname/workspaceblobstore for project workspace.
    :type storage_connections: Optional[List[str]]
    :param capability_host_kind: The kind of capability host, either as a string or CapabilityHostKind enum.
        Default is AGENTS.
    :type capability_host_kind: Union[str, CapabilityHostKind]
    :param kwargs: Additional keyword arguments.
    :type kwargs: Any

    .. admonition:: Example:

        .. literalinclude:: ../samples/ml_samples_capability_host.py
            :start-after: [START capability_host_object_create]
            :end-before: [END capability_host_object_create]
            :language: python
            :dedent: 8
            :caption: Create a CapabilityHost object.
    """

    def __init__(
        self,
        *,
        name: str,
        description: Optional[str] = None,
        vector_store_connections: Optional[List[str]] = None,
        ai_services_connections: Optional[List[str]] = None,
        storage_connections: Optional[List[str]] = None,
        thread_storage_connections: Optional[List[str]] = None,
        capability_host_kind: Union[str, CapabilityHostKind] = CapabilityHostKind.AGENTS,
        **kwargs: Any,
    ):
        super().__init__(name=name, description=description, **kwargs)
        self.capability_host_kind = capability_host_kind
        self.ai_services_connections = ai_services_connections
        self.storage_connections = storage_connections
        self.vector_store_connections = vector_store_connections
        self.thread_storage_connections = thread_storage_connections

    def dump(
        self,
        dest: Optional[Union[str, PathLike, IO[AnyStr]]],
        **kwargs: Any,
    ) -> None:
        """Dump the CapabilityHost content into a file in yaml format.

        :param dest: The destination to receive this CapabilityHost's content.
            Must be either a path to a local file, or an already-open file stream.
            If dest is a file path, a new file will be created,
            and an exception is raised if the file exists.
            If dest is an open file, the file will be written to directly,
            and an exception will be raised if the file is not writable.
        :type dest: Union[PathLike, str, IO[AnyStr]]
        """
        path = kwargs.pop("path", None)
        yaml_serialized = self._to_dict()
        dump_yaml_to_file(dest, yaml_serialized, default_flow_style=False, path=path, **kwargs)

    def _to_dict(self) -> Dict:
        """Dump the object into a dictionary.

        :return: Dictionary representation of the object.
        :rtype: Dict
        """

        return CapabilityHostSchema(context={BASE_PATH_CONTEXT_KEY: "./"}).dump(self)

    @classmethod
    def _load(
        cls,
        data: Optional[dict] = None,
        yaml_path: Optional[Union[os.PathLike, str]] = None,
        params_override: Optional[list] = None,
        **kwargs: Any,
    ) -> "CapabilityHost":
        """Load a capabilityhost object from a yaml file.

        :param cls: Indicates that this is a class method.
        :type cls: class
        :param data: Data Dictionary, defaults to None
        :type data: Dict
        :param yaml_path: YAML Path, defaults to None
        :type yaml_path: Union[PathLike, str]
        :param params_override: Fields to overwrite on top of the yaml file.
            Format is [{"field1": "value1"}, {"field2": "value2"}], defaults to None
        :type params_override: List[Dict]
        :raises Exception: An exception
        :return: Loaded CapabilityHost object.
        :rtype: ~azure.ai.ml.entities._workspace._ai_workspaces.capability_host.CapabilityHost
        """
        params_override = params_override or []
        data = data or {}
        context = {
            BASE_PATH_CONTEXT_KEY: Path(yaml_path).parent if yaml_path else Path("./"),
            PARAMS_OVERRIDE_KEY: params_override,
        }
        return cls(**load_from_dict(CapabilityHostSchema, data, context, **kwargs))

    @classmethod
    def _from_rest_object(cls, rest_obj: RestCapabilityHost) -> "CapabilityHost":
        """Convert a REST object into a CapabilityHost object.

        :param cls: Indicates that this is a class method.
        :type cls: class
        :param rest_obj: The REST object to convert.
        :type rest_obj: ~azure.ai.ml._restclient.v2024_10_01_preview.models._models_py3.CapabilityHost
        :return: CapabilityHost object.
        :rtype: ~azure.ai.ml.entities._workspace._ai_workspaces.capability_host.CapabilityHost
        """
        capability_host = cls(
            name=str(rest_obj.name),
            description=(rest_obj.properties.description if rest_obj.properties else None),
            ai_services_connections=(rest_obj.properties.ai_services_connections if rest_obj.properties else None),
            storage_connections=(rest_obj.properties.storage_connections if rest_obj.properties else None),
            vector_store_connections=(rest_obj.properties.vector_store_connections if rest_obj.properties else None),
            thread_storage_connections=(
                rest_obj.properties.thread_storage_connections if rest_obj.properties else None
            ),
            capability_host_kind=(
                rest_obj.properties.capability_host_kind if rest_obj.properties else CapabilityHostKind.AGENTS
            ),
        )
        return capability_host

    def _to_rest_object(self) -> RestCapabilityHost:
        """
        Convert the CapabilityHost instance to a RestCapabilityHost object.

        :return: A RestCapabilityHost object representing the capability host for a Hub or Project workspace.
        :rtype: azure.ai.ml._restclient.v2024_10_01_preview.models._models_py3.CapabilityHost
        """

        properties = RestCapabilityHostProperties(
            ai_services_connections=self.ai_services_connections,
            storage_connections=self.storage_connections,
            vector_store_connections=self.vector_store_connections,
            thread_storage_connections=self.thread_storage_connections,
            description=self.description,
            capability_host_kind=self.capability_host_kind,
        )
        resource = RestCapabilityHost(
            properties=properties,
        )
        return resource



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------
import logging

from marshmallow import fields

from azure.ai.ml._schema.core.fields import ArmStr, ComputeField, NestedField, UnionField
from azure.ai.ml._schema.core.resource import ResourceSchema
from azure.ai.ml._schema.job.identity import AMLTokenIdentitySchema, ManagedIdentitySchema, UserIdentitySchema
from azure.ai.ml.constants._common import AzureMLResourceType

from .creation_context import CreationContextSchema
from .services import (
    JobServiceSchema,
    SshJobServiceSchema,
    VsCodeJobServiceSchema,
    TensorBoardJobServiceSchema,
    JupyterLabJobServiceSchema,
)

module_logger = logging.getLogger(__name__)


class BaseJobSchema(ResourceSchema):
    creation_context = NestedField(CreationContextSchema, dump_only=True)
    services = fields.Dict(
        keys=fields.Str(),
        values=UnionField(
            [
                NestedField(SshJobServiceSchema),
                NestedField(TensorBoardJobServiceSchema),
                NestedField(VsCodeJobServiceSchema),
                NestedField(JupyterLabJobServiceSchema),
                # JobServiceSchema should be the last in the list.
                # To support types not set by users like Custom, Tracking, Studio.
                NestedField(JobServiceSchema),
            ],
            is_strict=True,
        ),
    )
    name = fields.Str()
    id = ArmStr(azureml_type=AzureMLResourceType.JOB, dump_only=True, required=False)
    display_name = fields.Str(required=False)
    tags = fields.Dict(keys=fields.Str(), values=fields.Str(allow_none=True))
    status = fields.Str(dump_only=True)
    experiment_name = fields.Str()
    properties = fields.Dict(keys=fields.Str(), values=fields.Str(allow_none=True))
    description = fields.Str()
    log_files = fields.Dict(
        keys=fields.Str(),
        values=fields.Str(),
        dump_only=True,
        metadata={
            "description": (
                "The list of log files associated with this run. This section is only populated "
                "by the service and will be ignored if contained in a yaml sent to the service "
                "(e.g. via `az ml job create` ...)"
            )
        },
    )
    compute = ComputeField(required=False)
    identity = UnionField(
        [
            NestedField(ManagedIdentitySchema),
            NestedField(AMLTokenIdentitySchema),
            NestedField(UserIdentitySchema),
        ]
    )



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

import logging
from typing import List

from azure.ai.ml._restclient.model_dataplane import AzureMachineLearningWorkspaces as ServiceClientModelDataplane
from azure.ai.ml._restclient.model_dataplane.models import BatchGetResolvedUrisDto, BatchModelPathResponseDto
from azure.ai.ml._scope_dependent_operations import OperationConfig, OperationScope, _ScopeDependentOperations

module_logger = logging.getLogger(__name__)


class ModelDataplaneOperations(_ScopeDependentOperations):
    def __init__(
        self,
        operation_scope: OperationScope,
        operation_config: OperationConfig,
        service_client: ServiceClientModelDataplane,
    ):
        super().__init__(operation_scope, operation_config)
        self._operation = service_client.models

    def get_batch_model_uris(self, model_ids: List[str]) -> BatchModelPathResponseDto:
        batch_uri_request = BatchGetResolvedUrisDto(values=model_ids)
        return self._operation.batch_get_resolved_uris(
            self._operation_scope.subscription_id,
            self._operation_scope.resource_group_name,
            self._workspace_name,
            body=batch_uri_request,
        )



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

# pylint: disable=redefined-builtin,disable=unused-argument

from typing import Any, Dict, Optional

from azure.ai.ml.entities._feature_set.source_process_code_metadata import SourceProcessCodeMetadata
from azure.ai.ml.exceptions import ErrorCategory, ErrorTarget, ValidationErrorType, ValidationException

from .delay_metadata import DelayMetadata
from .timestamp_column_metadata import TimestampColumnMetadata


class SourceMetadata(object):
    def __init__(
        self,
        *,
        type: str,
        timestamp_column: Optional[TimestampColumnMetadata] = None,
        path: Optional[str] = None,
        source_delay: Optional[DelayMetadata] = None,
        source_process_code: Optional[SourceProcessCodeMetadata] = None,
        dict: Optional[Dict] = None,
        **kwargs: Any,
    ):
        if type == "custom":
            # For custom feature source
            # Required: timestamp_column, dict and source_process_code.
            # Not support: path.
            if path:
                self.throw_exception("path", type, should_provide=False)
            if not (timestamp_column and dict and source_process_code):
                self.throw_exception("timestamp_column/dict/source_process_code", type, should_provide=True)
        elif type == "featureset":
            # For featureset feature source
            # Required: path.
            # Not support: timestamp_column, source_delay and source_process_code.
            if timestamp_column or source_delay or source_process_code:
                self.throw_exception("timestamp_column/source_delay/source_process_code", type, should_provide=False)
            if not path:
                self.throw_exception("path", type, should_provide=True)
        else:
            # For other type feature source
            # Required: timestamp_column, path.
            # Not support: source_process_code, dict
            if dict or source_process_code:
                self.throw_exception("dict/source_process_code", type, should_provide=False)
            if not (timestamp_column and path):
                self.throw_exception("timestamp_column/path", type, should_provide=True)
        self.type = type
        self.path = path
        self.timestamp_column = timestamp_column
        self.source_delay = source_delay
        self.source_process_code = source_process_code
        self.kwargs = dict

    @staticmethod
    def throw_exception(property_names: str, type: str, should_provide: bool):
        should_or_not = "need to" if should_provide else "cannot"
        msg = f"You {should_or_not} provide {property_names} for {type} feature source."
        raise ValidationException(
            message=msg,
            no_personal_data_message=msg,
            error_type=ValidationErrorType.INVALID_VALUE,
            target=ErrorTarget.FEATURE_SET,
            error_category=ErrorCategory.USER_ERROR,
        )



# coding=utf-8
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# Code generated by Microsoft (R) AutoRest Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is regenerated.
# --------------------------------------------------------------------------

from ._code_containers_operations import CodeContainersOperations
from ._code_versions_operations import CodeVersionsOperations
from ._component_containers_operations import ComponentContainersOperations
from ._component_versions_operations import ComponentVersionsOperations
from ._data_containers_operations import DataContainersOperations
from ._data_versions_operations import DataVersionsOperations
from ._data_references_operations import DataReferencesOperations
from ._environment_containers_operations import EnvironmentContainersOperations
from ._environment_versions_operations import EnvironmentVersionsOperations
from ._resource_management_asset_reference_operations import ResourceManagementAssetReferenceOperations
from ._model_containers_operations import ModelContainersOperations
from ._model_versions_operations import ModelVersionsOperations
from ._temporary_data_references_operations import TemporaryDataReferencesOperations

__all__ = [
    'CodeContainersOperations',
    'CodeVersionsOperations',
    'ComponentContainersOperations',
    'ComponentVersionsOperations',
    'DataContainersOperations',
    'DataVersionsOperations',
    'DataReferencesOperations',
    'EnvironmentContainersOperations',
    'EnvironmentVersionsOperations',
    'ResourceManagementAssetReferenceOperations',
    'ModelContainersOperations',
    'ModelVersionsOperations',
    'TemporaryDataReferencesOperations',
]



from typing import Tuple

import pytest
from test_utilities.utils import download_dataset

from azure.ai.ml.constants._common import AssetTypes
from azure.ai.ml.entities._inputs_outputs import Input
from azure.ai.ml.entities._job.automl.tabular.forecasting_settings import ForecastingSettings

TEST_CONFIG_PATH = "tests/test_configs/automl_job"
TEST_DATASETS_PATH = TEST_CONFIG_PATH + "/" + "test_datasets"

BANK_MARKETING_TRAIN_DATASET_PATH = TEST_DATASETS_PATH + "/" + "bank_marketing/train"
BANK_MARKETING_TEST_DATASET_PATH = TEST_DATASETS_PATH + "/" + "bank_marketing/valid"

BEER_FORECASTING_TRAIN_DATASET_PATH = TEST_DATASETS_PATH + "/" + "beer_forecasting/train"

CONLL_TRAIN_DATASET_PATH = TEST_DATASETS_PATH + "/" + "conll2003/train"
CONLL_VALID_DATASET_PATH = TEST_DATASETS_PATH + "/" + "conll2003/valid"

CREDIT_CARD_FRAUD_TRAIN_DATASET_PATH = TEST_DATASETS_PATH + "/" + "credit_card_fraud/train"
CREDIT_CARD_FRAUD_TEST_DATASET_PATH = TEST_DATASETS_PATH + "/" + "credit_card_fraud/test"
CREDIT_CARD_FRAUD_VALID_DATASET_PATH = TEST_DATASETS_PATH + "/" + "credit_card_fraud/valid"

MACHINE_DATA_TRAIN_DATASET_PATH = TEST_DATASETS_PATH + "/" + "machine_data/train"

NEWSGROUP_TRAIN_DATASET_PATH = TEST_DATASETS_PATH + "/" + "newsgroup/train"
NEWSGROUP_VALID_DATASET_PATH = TEST_DATASETS_PATH + "/" + "newsgroup/valid"

PAPER_CATEGORIZATION_TRAIN_DATASET_PATH = TEST_DATASETS_PATH + "/" + "paper_categorization/train"
PAPER_CATEGORIZATION_VALID_DATASET_PATH = TEST_DATASETS_PATH + "/" + "paper_categorization/valid"

IMAGE_CLASSIFICATION_TRAIN_DATASET_PATH = TEST_DATASETS_PATH + "/" + "image_classification/train"
IMAGE_CLASSIFICATION_VALID_DATASET_PATH = TEST_DATASETS_PATH + "/" + "image_classification/valid"

IMAGE_CLASSIFICATION_MULTILABEL_TRAIN_DATASET_PATH = TEST_DATASETS_PATH + "/" + "image_classification_multilabel/train"
IMAGE_CLASSIFICATION_MULTILABEL_VALID_DATASET_PATH = TEST_DATASETS_PATH + "/" + "image_classification_multilabel/valid"

IMAGE_OBJECT_DETECTION_TRAIN_DATASET_PATH = TEST_DATASETS_PATH + "/" + "image_object_detection/train"
IMAGE_OBJECT_DETECTION_VALID_DATASET_PATH = TEST_DATASETS_PATH + "/" + "image_object_detection/valid"

IMAGE_SEGMENTATION_TRAIN_DATASET_PATH = TEST_DATASETS_PATH + "/" + "image_instance_segmentation/train"
IMAGE_SEGMENTATION_VALID_DATASET_PATH = TEST_DATASETS_PATH + "/" + "image_instance_segmentation/valid"


@pytest.fixture
def bankmarketing_dataset() -> Tuple[Input, Input, str]:
    # Classification Dataset
    training_data = Input(type=AssetTypes.MLTABLE, path=BANK_MARKETING_TRAIN_DATASET_PATH)
    validation_data = Input(type=AssetTypes.MLTABLE, path=BANK_MARKETING_TEST_DATASET_PATH)
    label_column_name = "y"
    return training_data, validation_data, label_column_name


@pytest.fixture
def credit_card_fraud_dataset() -> Tuple[Input, Input, str]:
    # Classification Dataset
    training_data = Input(type=AssetTypes.MLTABLE, path=CREDIT_CARD_FRAUD_TRAIN_DATASET_PATH)
    validation_data = Input(type=AssetTypes.MLTABLE, path=CREDIT_CARD_FRAUD_VALID_DATASET_PATH)
    label_column_name = "Class"
    return training_data, validation_data, label_column_name


@pytest.fixture
def machinedata_dataset() -> Tuple[Input, str]:
    # Regression Dataset
    training_data = Input(type=AssetTypes.MLTABLE, path=MACHINE_DATA_TRAIN_DATASET_PATH)
    label_column_name = "ERP"
    return training_data, label_column_name


@pytest.fixture
def beer_forecasting_dataset() -> Tuple[Input, ForecastingSettings, str]:
    # Forecasting Dataset
    training_data = Input(type=AssetTypes.MLTABLE, path=BEER_FORECASTING_TRAIN_DATASET_PATH)
    label_column_name = "BeerProduction"
    time_column_name = "DATE"
    forecast_horizon = 12
    frequency = "MS"
    forecasting_settings = ForecastingSettings(
        time_column_name=time_column_name, forecast_horizon=forecast_horizon, frequency=frequency
    )

    return training_data, forecasting_settings, label_column_name


@pytest.fixture
def image_classification_dataset() -> Tuple[str, str]:
    # Download data
    download_url = "https://cvbp-secondary.z19.web.core.windows.net/datasets/image_classification/fridgeObjects.zip"
    data_file = "./fridgeObjects.zip"
    download_dataset(download_url=download_url, data_file=data_file)

    # Classification dataset MLTable paths
    return IMAGE_CLASSIFICATION_TRAIN_DATASET_PATH, IMAGE_CLASSIFICATION_VALID_DATASET_PATH


@pytest.fixture
def image_classification_multilabel_dataset() -> Tuple[str, str]:
    # Download data
    download_url = (
        "https://cvbp-secondary.z19.web.core.windows.net/datasets/image_classification/multilabelFridgeObjects.zip"
    )
    data_file = "./multilabelFridgeObjects.zip"
    download_dataset(download_url=download_url, data_file=data_file)

    # Multilabel classification dataset MLTable paths
    return IMAGE_CLASSIFICATION_MULTILABEL_TRAIN_DATASET_PATH, IMAGE_CLASSIFICATION_MULTILABEL_VALID_DATASET_PATH


@pytest.fixture
def image_object_detection_dataset() -> Tuple[str, str]:
    # Download data
    download_url = "https://cvbp-secondary.z19.web.core.windows.net/datasets/object_detection/odFridgeObjects.zip"
    data_file = "./odFridgeObjects.zip"
    download_dataset(download_url=download_url, data_file=data_file)

    # Image Object Dataset
    return IMAGE_OBJECT_DETECTION_TRAIN_DATASET_PATH, IMAGE_OBJECT_DETECTION_VALID_DATASET_PATH


@pytest.fixture
def image_segmentation_dataset() -> Tuple[str, str]:
    # Download data
    download_url = "https://cvbp-secondary.z19.web.core.windows.net/datasets/object_detection/odFridgeObjectsMask.zip"
    data_file = "./odFridgeObjectsMask.zip"
    download_dataset(download_url=download_url, data_file=data_file)

    # Image Object Dataset
    return IMAGE_SEGMENTATION_TRAIN_DATASET_PATH, IMAGE_SEGMENTATION_VALID_DATASET_PATH


# Text Classification Dataset
@pytest.fixture
def newsgroup() -> Tuple[Input, Input, str]:
    training_data = Input(type=AssetTypes.MLTABLE, path=NEWSGROUP_TRAIN_DATASET_PATH)
    validation_data = Input(type=AssetTypes.MLTABLE, path=NEWSGROUP_VALID_DATASET_PATH)
    target_column_name = "y"

    return training_data, validation_data, target_column_name


# Text Classification Multilabel Dataset
@pytest.fixture
def paper_categorization() -> Tuple[Input, Input, str]:
    training_data = Input(type=AssetTypes.MLTABLE, path=PAPER_CATEGORIZATION_TRAIN_DATASET_PATH)
    validation_data = Input(type=AssetTypes.MLTABLE, path=PAPER_CATEGORIZATION_VALID_DATASET_PATH)
    target_column_name = "terms"

    return training_data, validation_data, target_column_name


# Text NER Dataset
@pytest.fixture
def conll() -> Tuple[Input, Input]:
    training_data = Input(type=AssetTypes.MLTABLE, path=CONLL_TRAIN_DATASET_PATH)
    validation_data = Input(type=AssetTypes.MLTABLE, path=CONLL_VALID_DATASET_PATH)

    return training_data, validation_data



import os
import tempfile
from collections import OrderedDict

import pytest

from azure.ai.ml._utils.utils import (
    _get_mfe_base_url_from_batch_endpoint,
    dict_eq,
    get_all_data_binding_expressions,
    is_data_binding_expression,
    map_single_brackets_and_warn,
    write_to_shared_file,
    get_valid_dot_keys_with_wildcard,
)
from azure.ai.ml.entities import BatchEndpoint
from azure.ai.ml.entities._util import convert_ordered_dict_to_dict


@pytest.mark.unittest
@pytest.mark.core_sdk_test
class TestUtils:
    def test_dict_eq(self) -> None:
        assert dict_eq(dict1={}, dict2=None)
        assert dict_eq(dict1={"blue": 1, "green": 2}, dict2={"blue": 1, "green": 2})

    def test_get_mfe_base_url_from_batch_endpoint(self) -> None:
        endpoint = BatchEndpoint(
            name="bla",
            scoring_uri="https://eastus2euap.api.azureml.ms/mferp/managementfrontend/subscriptions/4faaaf21-663f-4391-96fd-47197c630979/resourceGroups/test-rg-eastus2euap-v2-2021W16/providers/Microsoft.MachineLearningServices/workspaces/sdk_vnext_cli/batchEndpoints/be-20210421/jobs?api-version=2020-09-01-preview",
        )
        assert "https://eastus2euap.api.azureml.ms/mferp/managementfrontend" == _get_mfe_base_url_from_batch_endpoint(
            endpoint
        )

    def test_legacy_command_substitution(self) -> None:
        input_string = "test {inputs.val1} ${{inputs.val2}} {search_space.val3} ${{search_space.val4}} {non-captured}"
        expected_string = (
            "test ${{inputs.val1}} ${{inputs.val2}} ${{search_space.val3}} ${{search_space.val4}} {non-captured}"
        )

        output_string = map_single_brackets_and_warn(input_string)

        assert expected_string == output_string

    def test_is_data_binding_expression(self) -> None:
        target_string = "${{  parent.inputs.input_string  }}"
        assert is_data_binding_expression(target_string)
        assert is_data_binding_expression(target_string, ["parent", "inputs", "input_string"])
        assert not is_data_binding_expression(target_string, ["parent", "outputs"])
        assert not is_data_binding_expression(target_string, ["parent", "inputs", "input_string_2"])

        target_string = "${{  parent.inputs.input_string  }} abc ${{  parent.outputs.output_string  }}"
        assert not is_data_binding_expression(target_string)
        assert is_data_binding_expression(target_string, is_singular=False)
        assert is_data_binding_expression(target_string, ["parent", "inputs", "input_string"], is_singular=False)
        assert not is_data_binding_expression(target_string, ["parent", "name"], is_singular=False)

    def test_get_all_data_binding_expression(self):
        target_string = "cat ${{inputs.input_folder}}/${{inputs.file_name}} && cp ${{inputs.input_folder}}/${{inputs.file_name}} ${{outputs.output_folder}}/${{inputs.file_name}}"
        data_binding_expressions = get_all_data_binding_expressions(target_string, is_singular=False)
        assert data_binding_expressions == [
            "inputs.input_folder",
            "inputs.file_name",
            "inputs.input_folder",
            "inputs.file_name",
            "outputs.output_folder",
            "inputs.file_name",
        ]

    def test_ordered_dict_nested_in_list_conversion(self):
        ordered_dict1, ordered_dict2 = OrderedDict(), OrderedDict()
        ordered_dict1["a"] = 1
        ordered_dict1["b"] = 2
        ordered_dict2["d"] = 3
        ordered_dict2["c"] = 4
        target_list = [ordered_dict1, ordered_dict2]
        assert convert_ordered_dict_to_dict(target_list) == [{"a": 1, "b": 2}, {"d": 3, "c": 4}]

    def test_open_with_int_mode(self):
        def get_int_mode(file_path: str) -> str:
            int_mode = os.stat(file_path).st_mode & 0o777
            return oct(int_mode)

        with tempfile.TemporaryDirectory() as temp_dir:
            target_file_path = temp_dir + "/test.txt"
            with open(target_file_path, "w") as f:
                # default mode is 0o666 for windows and 0o755 for linux
                f.write("test")
            write_to_shared_file(target_file_path, "test2")
            # check that the file mode is preserved
            assert get_int_mode(target_file_path) == "0o666"
            with open(target_file_path, "r") as f:
                assert f.read() == "test2"

    def test_get_valid_dot_keys_with_wildcard(self):
        root = {
            "simple": 1,
            "deep": {
                "l1": {
                    "l2": 1,
                    "l2_2": 2,
                },
                "l1_2": {
                    "l2": 3,
                },
            },
        }
        assert get_valid_dot_keys_with_wildcard(root, "simple") == ["simple"]
        assert get_valid_dot_keys_with_wildcard(root, "deep.l1.l2") == ["deep.l1.l2"]
        assert get_valid_dot_keys_with_wildcard(root, "deep.*.l2") == ["deep.l1.l2", "deep.l1_2.l2"]
        assert get_valid_dot_keys_with_wildcard(root, "deep.*.*") == ["deep.l1.l2", "deep.l1.l2_2", "deep.l1_2.l2"]
        assert get_valid_dot_keys_with_wildcard(root, "deep.*.l2_2") == ["deep.l1.l2_2"]
        assert get_valid_dot_keys_with_wildcard(root, "deep.*.l2_3") == []
        assert get_valid_dot_keys_with_wildcard(root, "deep.*.l2.*") == []
        assert get_valid_dot_keys_with_wildcard(root, "deep.*.l2.*.l3") == []

        assert get_valid_dot_keys_with_wildcard(
            root,
            "deep.*.*",
            validate_func=lambda _root, _parts: _parts[1] == "l1_2",
        ) == ["deep.l1_2.l2"]



# coding: utf-8
"""Comparison of `binary` and `xentropy` objectives.

BLUF: The `xentropy` objective does logistic regression and generalizes
to the case where labels are probabilistic (i.e. numbers between 0 and 1).

Details: Both `binary` and `xentropy` minimize the log loss and use
`boost_from_average = TRUE` by default. Possibly the only difference
between them with default settings is that `binary` may achieve a slight
speed improvement by assuming that the labels are binary instead of
probabilistic.
"""

import time

import lightgbm as lgb
import numpy as np
import pandas as pd
from scipy.special import expit

#################
# Simulate some binary data with a single categorical and
#   single continuous predictor
np.random.seed(0)
N = 1000
X = pd.DataFrame({"continuous": range(N), "categorical": np.repeat([0, 1, 2, 3, 4], N / 5)})
CATEGORICAL_EFFECTS = [-1, -1, -2, -2, 2]
LINEAR_TERM = np.array(
    [-0.5 + 0.01 * X["continuous"][k] + CATEGORICAL_EFFECTS[X["categorical"][k]] for k in range(X.shape[0])]
) + np.random.normal(0, 1, X.shape[0])
TRUE_PROB = expit(LINEAR_TERM)
Y = np.random.binomial(1, TRUE_PROB, size=N)
DATA = {
    "X": X,
    "probability_labels": TRUE_PROB,
    "binary_labels": Y,
    "lgb_with_binary_labels": lgb.Dataset(X, Y),
    "lgb_with_probability_labels": lgb.Dataset(X, TRUE_PROB),
}


#################
# Set up a couple of utilities for our experiments
def log_loss(preds, labels):
    """Logarithmic loss with non-necessarily-binary labels."""
    log_likelihood = np.sum(labels * np.log(preds)) / len(preds)
    return -log_likelihood


def experiment(objective, label_type, data):
    """Measure performance of an objective.

    Parameters
    ----------
    objective : string 'binary' or 'xentropy'
        Objective function.
    label_type : string 'binary' or 'probability'
        Type of the label.
    data : dict
        Data for training.

    Returns
    -------
    result : dict
        Experiment summary stats.
    """
    np.random.seed(0)
    nrounds = 5
    lgb_data = data["lgb_with_" + label_type + "_labels"]
    params = {"objective": objective, "feature_fraction": 1, "bagging_fraction": 1, "verbose": -1}
    time_zero = time.time()
    gbm = lgb.train(params, lgb_data, num_boost_round=nrounds)
    y_fitted = gbm.predict(data["X"])
    y_true = data[label_type + "_labels"]
    duration = time.time() - time_zero
    return {"time": duration, "correlation": np.corrcoef(y_fitted, y_true)[0, 1], "logloss": log_loss(y_fitted, y_true)}


#################
# Observe the behavior of `binary` and `xentropy` objectives
print("Performance of `binary` objective with binary labels:")
print(experiment("binary", label_type="binary", data=DATA))

print("Performance of `xentropy` objective with binary labels:")
print(experiment("xentropy", label_type="binary", data=DATA))

print("Performance of `xentropy` objective with probability labels:")
print(experiment("xentropy", label_type="probability", data=DATA))

# Trying this throws an error on non-binary values of y:
#   experiment('binary', label_type='probability', DATA)

# The speed of `binary` is not drastically different than
#   `xentropy`. `xentropy` runs faster than `binary` in many cases, although
#   there are reasons to suspect that `binary` should run faster when the
#   label is an integer instead of a float
K = 10
A = [experiment("binary", label_type="binary", data=DATA)["time"] for k in range(K)]
B = [experiment("xentropy", label_type="binary", data=DATA)["time"] for k in range(K)]
print("Best `binary` time: " + str(min(A)))
print("Best `xentropy` time: " + str(min(B)))



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

# pylint: disable=protected-access

from typing import TYPE_CHECKING, Any, Dict, Optional, Union

from azure.ai.ml._restclient.v2023_04_01_preview.models import AutoMLJob as RestAutoMLJob
from azure.ai.ml._restclient.v2023_04_01_preview.models import JobBase, TaskType
from azure.ai.ml._restclient.v2023_04_01_preview.models._azure_machine_learning_workspaces_enums import (
    ClassificationPrimaryMetrics,
)
from azure.ai.ml._restclient.v2024_01_01_preview.models import TextNer as RestTextNER
from azure.ai.ml._utils.utils import camel_to_snake, is_data_binding_expression
from azure.ai.ml.constants._common import BASE_PATH_CONTEXT_KEY
from azure.ai.ml.constants._job.automl import AutoMLConstants
from azure.ai.ml.entities._credentials import _BaseJobIdentityConfiguration
from azure.ai.ml.entities._inputs_outputs import Input
from azure.ai.ml.entities._job._input_output_helpers import from_rest_data_outputs, to_rest_data_outputs
from azure.ai.ml.entities._job.automl.nlp.automl_nlp_job import AutoMLNLPJob
from azure.ai.ml.entities._job.automl.nlp.nlp_featurization_settings import NlpFeaturizationSettings
from azure.ai.ml.entities._job.automl.nlp.nlp_fixed_parameters import NlpFixedParameters
from azure.ai.ml.entities._job.automl.nlp.nlp_limit_settings import NlpLimitSettings
from azure.ai.ml.entities._job.automl.nlp.nlp_sweep_settings import NlpSweepSettings
from azure.ai.ml.entities._system_data import SystemData
from azure.ai.ml.entities._util import load_from_dict

# avoid circular import error
if TYPE_CHECKING:
    from azure.ai.ml.entities._component.component import Component


class TextNerJob(AutoMLNLPJob):
    """Configuration for AutoML Text NER Job.

    :param training_data: Training data to be used for training, defaults to None
    :type training_data: Optional[~azure.ai.ml.Input]
    :param validation_data: Validation data to be used for evaluating the trained model,
        defaults to None
    :type validation_data: Optional[~azure.ai.ml.Input]
    :param primary_metric: The primary metric to be displayed, defaults to None
    :type primary_metric: Optional[str]
    :param log_verbosity: Log verbosity level, defaults to None
    :type log_verbosity: Optional[str]

    .. admonition:: Example:

        .. literalinclude:: ../samples/ml_samples_automl_nlp.py
                :start-after: [START automl.text_ner_job]
                :end-before: [END automl.text_ner_job]
                :language: python
                :dedent: 8
                :caption: creating an automl text ner job
    """

    _DEFAULT_PRIMARY_METRIC = ClassificationPrimaryMetrics.ACCURACY

    def __init__(
        self,
        *,
        training_data: Optional[Input] = None,
        validation_data: Optional[Input] = None,
        primary_metric: Optional[str] = None,
        log_verbosity: Optional[str] = None,
        **kwargs: Any
    ):
        super(TextNerJob, self).__init__(
            task_type=TaskType.TEXT_NER,
            primary_metric=primary_metric or TextNerJob._DEFAULT_PRIMARY_METRIC,
            training_data=training_data,
            validation_data=validation_data,
            log_verbosity=log_verbosity,
            **kwargs,
        )

    @property
    def primary_metric(self) -> Union[str, ClassificationPrimaryMetrics]:
        return self._primary_metric

    @primary_metric.setter
    def primary_metric(self, value: Union[str, ClassificationPrimaryMetrics]) -> None:
        if is_data_binding_expression(str(value), ["parent"]):
            self._primary_metric = value
            return

        self._primary_metric = (
            TextNerJob._DEFAULT_PRIMARY_METRIC
            if value is None
            else ClassificationPrimaryMetrics[camel_to_snake(value).upper()]
        )

    def _to_rest_object(self) -> JobBase:
        text_ner = RestTextNER(
            training_data=self.training_data,
            validation_data=self.validation_data,
            limit_settings=self._limits._to_rest_object() if self._limits else None,
            sweep_settings=self._sweep._to_rest_object() if self._sweep else None,
            fixed_parameters=self._training_parameters._to_rest_object() if self._training_parameters else None,
            search_space=(
                [entry._to_rest_object() for entry in self._search_space if entry is not None]
                if self._search_space is not None
                else None
            ),
            featurization_settings=self._featurization._to_rest_object() if self._featurization else None,
            primary_metric=self.primary_metric,
            log_verbosity=self.log_verbosity,
        )
        # resolve data inputs in rest object
        self._resolve_data_inputs(text_ner)

        properties = RestAutoMLJob(
            display_name=self.display_name,
            description=self.description,
            experiment_name=self.experiment_name,
            tags=self.tags,
            compute_id=self.compute,
            properties=self.properties,
            environment_id=self.environment_id,
            environment_variables=self.environment_variables,
            services=self.services,
            outputs=to_rest_data_outputs(self.outputs),
            resources=self.resources,
            task_details=text_ner,
            identity=self.identity._to_job_rest_object() if self.identity else None,
            queue_settings=self.queue_settings,
        )

        result = JobBase(properties=properties)
        result.name = self.name
        return result

    @classmethod
    def _from_rest_object(cls, obj: JobBase) -> "TextNerJob":
        properties: RestAutoMLJob = obj.properties
        task_details: RestTextNER = properties.task_details
        assert isinstance(task_details, RestTextNER)
        limits = (
            NlpLimitSettings._from_rest_object(task_details.limit_settings) if task_details.limit_settings else None
        )
        featurization = (
            NlpFeaturizationSettings._from_rest_object(task_details.featurization_settings)
            if task_details.featurization_settings
            else None
        )
        sweep = NlpSweepSettings._from_rest_object(task_details.sweep_settings) if task_details.sweep_settings else None
        training_parameters = (
            NlpFixedParameters._from_rest_object(task_details.fixed_parameters)
            if task_details.fixed_parameters
            else None
        )

        text_ner_job = cls(
            # ----- job specific params
            id=obj.id,
            name=obj.name,
            description=properties.description,
            tags=properties.tags,
            properties=properties.properties,
            experiment_name=properties.experiment_name,
            services=properties.services,
            status=properties.status,
            creation_context=SystemData._from_rest_object(obj.system_data) if obj.system_data else None,
            display_name=properties.display_name,
            compute=properties.compute_id,
            outputs=from_rest_data_outputs(properties.outputs),
            resources=properties.resources,
            # ----- task specific params
            primary_metric=task_details.primary_metric,
            log_verbosity=task_details.log_verbosity,
            target_column_name=task_details.target_column_name,
            training_data=task_details.training_data,
            validation_data=task_details.validation_data,
            limits=limits,
            sweep=sweep,
            training_parameters=training_parameters,
            search_space=cls._get_search_space_from_str(task_details.search_space),
            featurization=featurization,
            identity=(
                _BaseJobIdentityConfiguration._from_rest_object(properties.identity) if properties.identity else None
            ),
            queue_settings=properties.queue_settings,
        )

        text_ner_job._restore_data_inputs()

        return text_ner_job

    def _to_component(self, context: Optional[Dict] = None, **kwargs: Any) -> "Component":
        raise NotImplementedError()

    @classmethod
    def _load_from_dict(cls, data: Dict, context: Dict, additional_message: str, **kwargs: Any) -> "TextNerJob":
        from azure.ai.ml._schema.automl.nlp_vertical.text_ner import TextNerSchema

        if kwargs.pop("inside_pipeline", False):
            from azure.ai.ml._schema.pipeline.automl_node import AutoMLTextNerNode

            loaded_data = load_from_dict(AutoMLTextNerNode, data, context, additional_message, **kwargs)
        else:
            loaded_data = load_from_dict(TextNerSchema, data, context, additional_message, **kwargs)
        job_instance = cls._create_instance_from_schema_dict(loaded_data)
        return job_instance

    @classmethod
    def _create_instance_from_schema_dict(cls, loaded_data: Dict) -> "TextNerJob":
        loaded_data.pop(AutoMLConstants.TASK_TYPE_YAML, None)
        return TextNerJob(**loaded_data)

    def _to_dict(self, inside_pipeline: bool = False) -> Dict:
        from azure.ai.ml._schema.automl.nlp_vertical.text_ner import TextNerSchema
        from azure.ai.ml._schema.pipeline.automl_node import AutoMLTextNerNode

        if inside_pipeline:
            res_autoML: dict = AutoMLTextNerNode(context={BASE_PATH_CONTEXT_KEY: "./"}).dump(self)
            return res_autoML

        res: dict = TextNerSchema(context={BASE_PATH_CONTEXT_KEY: "./"}).dump(self)
        return res

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, TextNerJob):
            return NotImplemented

        if not super(TextNerJob, self).__eq__(other):
            return False

        return self.primary_metric == other.primary_metric

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

import abc
from typing import Any, Optional, cast

from .._exceptions import CredentialUnavailableError
from .._internal.get_token_mixin import GetTokenMixin
from .._internal.managed_identity_client import ManagedIdentityClient


class ManagedIdentityBase(GetTokenMixin):
    """Base class for internal credentials using ManagedIdentityClient."""

    from azure.core.credentials import AccessToken

    def __init__(self, **kwargs: Any):
        super(ManagedIdentityBase, self).__init__()
        self._client = self.get_client(**kwargs)

    @abc.abstractmethod
    def get_client(self, **kwargs: Any) -> Optional[ManagedIdentityClient]:
        pass

    @abc.abstractmethod
    def get_unavailable_message(self) -> str:
        pass

    def __enter__(self) -> "ManagedIdentityBase":
        if self._client:
            self._client.__enter__()
        return self

    def __exit__(self, *args: Any) -> None:
        if self._client:
            self._client.__exit__(*args)

    def close(self) -> None:
        self.__exit__()

    def get_token(self, *scopes: str, **kwargs: Any) -> AccessToken:
        if not self._client:
            raise CredentialUnavailableError(message=self.get_unavailable_message())
        return super(ManagedIdentityBase, self).get_token(*scopes, **kwargs)

    def _acquire_token_silently(self, *scopes: str, **kwargs: Any) -> Optional[AccessToken]:
        # casting because mypy can't determine that these methods are called
        # only by get_token, which raises when self._client is None
        return cast(ManagedIdentityClient, self._client).get_cached_token(*scopes)

    def _request_token(self, *scopes: str, **kwargs: Any) -> AccessToken:
        return cast(ManagedIdentityClient, self._client).request_token(*scopes, **kwargs)



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

import logging
import re
from typing import Optional, Tuple, Union

from azure.ai.ml._artifacts._blob_storage_helper import BlobStorageClient
from azure.ai.ml._artifacts._constants import STORAGE_URI_REGEX
from azure.ai.ml._artifacts._fileshare_storage_helper import FileStorageClient
from azure.ai.ml._artifacts._gen2_storage_helper import Gen2StorageClient
from azure.ai.ml._azure_environments import _get_storage_endpoint_from_metadata
from azure.ai.ml._restclient.v2022_10_01.models import DatastoreType
from azure.ai.ml.constants._common import (
    FILE_PREFIX,
    FOLDER_PREFIX,
    JOB_URI_REGEX_FORMAT,
    LONG_URI_FORMAT,
    LONG_URI_REGEX_FORMAT,
    MLFLOW_URI_REGEX_FORMAT,
    OUTPUT_URI_REGEX_FORMAT,
    SHORT_URI_FORMAT,
    SHORT_URI_REGEX_FORMAT,
    STORAGE_ACCOUNT_URLS,
)
from azure.ai.ml.exceptions import ErrorTarget, MlException, ValidationErrorType, ValidationException

module_logger = logging.getLogger(__name__)


class AzureMLDatastorePathUri:
    """Parser for an azureml:// datastore path URI, e.g.: azureml://datastores/mydatastore/paths/images/dogs'.

    :param uri: The AzureML datastore path URI.
    :type uri: str
    :raises ~azure.ai.ml.exceptions.ValidationException: Raised if the AzureML datastore
        path URI is incorrectly formatted.
    '
    """

    def __init__(self, uri: str):
        if uri.startswith(FILE_PREFIX):
            uri = uri[len(FILE_PREFIX) :]
        elif uri.startswith(FOLDER_PREFIX):
            uri = uri[len(FOLDER_PREFIX) :]
        self.uri = uri

        short_uri_match = re.match(SHORT_URI_REGEX_FORMAT, uri)
        ml_flow_uri_match = re.match(MLFLOW_URI_REGEX_FORMAT, uri)
        job_uri_match = re.match(JOB_URI_REGEX_FORMAT, uri)
        long_uri_match = re.match(LONG_URI_REGEX_FORMAT, uri)
        output_uri_match = re.match(OUTPUT_URI_REGEX_FORMAT, uri)

        if short_uri_match:
            self.datastore = short_uri_match.group(1)
            self.path = short_uri_match.group(2)
            self.uri_type = "Datastore"
            self.workspace_name = None
            self.resource_group = None
            self.subscription_id = None
        elif ml_flow_uri_match:
            self.datastore = ml_flow_uri_match.group(1)
            self.path = ml_flow_uri_match.group(2)
            self.uri_type = "MlFlow"
            self.workspace_name = None
            self.resource_group = None
            self.subscription_id = None
        elif job_uri_match:
            self.datastore = job_uri_match.group(1)
            self.path = job_uri_match.group(2)
            self.uri_type = "Job"
            self.workspace_name = None
            self.resource_group = None
            self.subscription_id = None
        elif output_uri_match:
            self.datastore = output_uri_match.group(1)
            self.path = output_uri_match.group(2)
            self.uri_type = None
            self.workspace_name = None
            self.resource_group = None
            self.subscription_id = None
        elif long_uri_match:
            self.datastore = long_uri_match.group(4)
            self.path = long_uri_match.group(5)
            self.uri_type = "Datastore"
            self.workspace_name = long_uri_match.group(3)
            self.resource_group = long_uri_match.group(2)
            self.subscription_id = long_uri_match.group(1)
        else:
            msg = "Invalid AzureML datastore path URI {}"
            raise ValidationException(
                message=msg.format(uri),
                no_personal_data_message=msg.format("[uri]"),
                target=ErrorTarget.DATASTORE,
                error_type=ValidationErrorType.INVALID_VALUE,
            )

    def to_short_uri(self) -> str:
        return SHORT_URI_FORMAT.format(self.datastore, self.path)

    def to_long_uri(self, subscription_id: str, resource_group_name: str, workspace_name: str) -> str:
        return LONG_URI_FORMAT.format(
            subscription_id,
            resource_group_name,
            workspace_name,
            self.datastore,
            self.path,
        )

    def get_uri_type(self) -> str:
        if self.uri[0:20] == "azureml://datastores":
            return "Datastore"
        if self.uri[0:14] == "azureml://jobs":
            return "Jobs"

        if self.uri[0 : self.uri.find(":")] == "runs":
            return "MLFlow"
        msg = "Invalid uri format for {}. URI must start with 'azureml://' or 'runs:/'"
        raise ValidationException(
            message=msg.format(self.uri),
            no_personal_data_message=msg.format("[self.uri]"),
            target=ErrorTarget.DATASTORE,
            error_type=ValidationErrorType.INVALID_VALUE,
        )


def get_storage_client(
    credential: str,
    storage_account: str,
    storage_type: Union[DatastoreType, str] = DatastoreType.AZURE_BLOB,
    account_url: Optional[str] = None,
    container_name: Optional[str] = None,
) -> Union[BlobStorageClient, FileStorageClient, Gen2StorageClient]:
    """Return a storage client class instance based on the storage account type.

    :param credential: The credential
    :type credential: str
    :param storage_account: The storage_account name
    :type storage_account: str
    :param storage_type: The storage type
    :type storage_type: Union[DatastoreType, str]
    :param account_url: The account url
    :type account_url: Optional[str]
    :param container_name: The container name
    :type container_name: Optional[str]
    :return: The storage client
    :rtype: Union[BlobStorageClient, FileStorageClient, Gen2StorageClient]
    """
    client_builders = {
        DatastoreType.AZURE_BLOB: lambda credential, container_name, account_url: BlobStorageClient(
            credential=credential, account_url=account_url, container_name=container_name
        ),
        DatastoreType.AZURE_DATA_LAKE_GEN2: lambda credential, container_name, account_url: Gen2StorageClient(
            credential=credential, file_system=container_name, account_url=account_url
        ),
        DatastoreType.AZURE_FILE: lambda credential, container_name, account_url: FileStorageClient(
            credential=credential, file_share_name=container_name, account_url=account_url
        ),
    }

    if storage_type not in client_builders:
        msg = (
            f"Datastore type {storage_type} is not supported. Supported storage"
            + f"types for artifact upload include: {*client_builders,}"
        )
        raise ValidationException(
            message=msg,
            no_personal_data_message=msg,
            target=ErrorTarget.DATASTORE,
            error_type=ValidationErrorType.INVALID_VALUE,
        )

    storage_endpoint = _get_storage_endpoint_from_metadata()
    if not account_url and storage_endpoint:
        account_url = STORAGE_ACCOUNT_URLS[storage_type].format(storage_account, storage_endpoint)
    return client_builders[storage_type](credential, container_name, account_url)


def get_artifact_path_from_storage_url(blob_url: str, container_name: dict) -> str:
    split_blob_url = blob_url.split(container_name)
    if len(split_blob_url) > 1:
        path = split_blob_url[-1]
        if path.startswith("/"):
            return path[1:]
        return path
    return blob_url


def get_ds_name_and_path_prefix(asset_uri: str, registry_name: Optional[str] = None) -> Tuple[str, str]:
    if registry_name:
        try:
            split_paths = re.findall(STORAGE_URI_REGEX, asset_uri)
            path_prefix = split_paths[0][-1]
        except Exception as e:
            msg = "Registry asset URI could not be parsed."
            raise MlException(message=msg, no_personal_data_message=msg) from e
        ds_name = None
    else:
        try:
            ds_name = asset_uri.split("paths")[0].split("/")[-2]
            path_prefix = asset_uri.split("paths")[1][1:]
        except Exception as e:
            msg = "Workspace asset URI could not be parsed."
            raise MlException(message=msg, no_personal_data_message=msg) from e

    return ds_name, path_prefix



import pytest

from azure.ai.ml import UserIdentityConfiguration
from azure.ai.ml.constants._common import AssetTypes
from azure.ai.ml.entities import CommandJob, CommandJobLimits, Job
from azure.ai.ml.entities._assets import Code
from azure.ai.ml.entities._builders.command_func import command
from azure.ai.ml.entities._inputs_outputs import Input, Output
from azure.ai.ml.entities._job.distribution import MpiDistribution, PyTorchDistribution, TensorFlowDistribution
from azure.ai.ml.entities._job.job_resource_configuration import JobResourceConfiguration
from azure.ai.ml.entities._job.sweep.early_termination_policy import TruncationSelectionPolicy
from azure.ai.ml.entities._job.sweep.objective import Objective
from azure.ai.ml.entities._job.sweep.search_space import LogUniform
from azure.ai.ml.sweep import (
    BayesianSamplingAlgorithm,
    Choice,
    GridSamplingAlgorithm,
    RandomSamplingAlgorithm,
    SweepJob,
    SweepJobLimits,
)


@pytest.mark.unittest
@pytest.mark.training_experiences_test
class TestSweepJob:
    def test_sweep_job_top_level_properties(self):
        command_job = CommandJob(
            code="./src",
            command="python train.py --ss {search_space.ss}",
            # Is it expected that using dicts do not work?
            # inputs={"input1": {"path": "trial.csv", "type": "uri_file", "mode": "rw_mount"}},
            # outputs={"default": {"mode": "rw_mount"}},
            inputs={"input1": Input(path="trial.csv")},
            outputs={"default": Output(path="./foo")},
            compute="trial",
            environment="AzureML-sklearn-1.0-ubuntu20.04-py38-cpu:33",
            limits=CommandJobLimits(timeout=120),
        )
        sweep = SweepJob(
            sampling_algorithm="random",
            trial=command_job,
            search_space={"ss": Choice(type="choice", values=[{"space1": True}, {"space2": True}])},
            inputs={"input1": {"path": "top_level.csv", "type": "uri_file", "mode": "ro_mount"}},
            compute="top_level",
            limits=SweepJobLimits(trial_timeout=600),
            identity=UserIdentityConfiguration(),
        )
        rest = sweep._to_rest_object()
        sweep_job: SweepJob = Job._from_rest_object(rest)

        assert sweep_job.compute == sweep.compute
        assert sweep_job.limits.trial_timeout == sweep.limits.trial_timeout
        assert sweep_job.inputs == sweep.inputs
        assert sweep_job.outputs == sweep.outputs
        assert sweep_job.identity == sweep.identity

    def test_sweep_job_override_missing_properties(self):
        command_job = CommandJob(
            code="./src",
            command="python train.py --ss {search_space.ss}",
            # Is it expected that using dicts do not work?
            # inputs={"input1": {"path": "trial.csv", "type": "uri_file", "mode": "rw_mount"}},
            # outputs={"default": {"mode": "rw_mount"}},
            inputs={"input1": Input(path="trial.csv")},
            outputs={"default": Output(path="./foo")},
            compute="trial",
            environment="AzureML-sklearn-1.0-ubuntu20.04-py38-cpu:33",
            limits=CommandJobLimits(timeout=120),
        )
        sweep = SweepJob(
            sampling_algorithm="random",
            trial=command_job,
            search_space={"ss": Choice(type="choice", values=[{"space1": True}, {"space2": True}])},
            inputs=None,
        )
        rest = sweep._to_rest_object()
        sweep_job: SweepJob = Job._from_rest_object(rest)

        assert sweep_job.compute == command_job.compute
        assert sweep_job.limits.trial_timeout == command_job.limits.timeout
        assert sweep_job.inputs == command_job.inputs
        assert sweep_job.outputs == command_job.outputs

    @pytest.mark.parametrize(
        "sampling_algorithm, expected_rest_type",
        [
            ("random", "Random"),
            ("grid", "Grid"),
            ("bayesian", "Bayesian"),
            (RandomSamplingAlgorithm(), "Random"),
            (GridSamplingAlgorithm(), "Grid"),
            (BayesianSamplingAlgorithm(), "Bayesian"),
        ],
    )
    def test_sampling_algorithm_to_rest(self, sampling_algorithm, expected_rest_type):
        command_job = CommandJob(
            code=Code(base_path="./src"),
            command="python train.py --ss {search_space.ss}",
            inputs={"input1": {"file": "trial.csv", "mode": "ro_mount"}},
            outputs={"default": {"mode": "rw_mount"}},
            compute="trial",
            environment="AzureML-sklearn-1.0-ubuntu20.04-py38-cpu:33",
            limits=CommandJobLimits(timeout=120),
        )

        sweep = SweepJob(
            sampling_algorithm=sampling_algorithm,
            trial=command_job,
            search_space={"ss": Choice(type="choice", values=[{"space1": True}, {"space2": True}])},
            inputs={"input1": {"file": "top_level.csv", "mode": "ro_mount"}},
            compute="top_level",
            limits=SweepJobLimits(trial_timeout=600),
        )

        rest = sweep._to_rest_object()
        assert rest.properties.sampling_algorithm.sampling_algorithm_type == expected_rest_type

    @pytest.mark.parametrize(
        "sampling_algorithm, expected_from_rest_type",
        [
            ("random", "random"),
            ("grid", "grid"),
            ("bayesian", "bayesian"),
            (RandomSamplingAlgorithm(), "random"),
            (GridSamplingAlgorithm(), "grid"),
            (BayesianSamplingAlgorithm(), "bayesian"),
        ],
    )
    def test_sampling_algorithm_from_rest(self, sampling_algorithm, expected_from_rest_type):
        command_job = CommandJob(
            code=Code(base_path="./src"),
            command="python train.py --ss {search_space.ss}",
            inputs={"input1": Input(path="trial.csv")},
            outputs={"default": Output(path="./foo")},
            compute="trial",
            environment="AzureML-sklearn-1.0-ubuntu20.04-py38-cpu:33",
            limits=CommandJobLimits(timeout=120),
        )

        sweep = SweepJob(
            sampling_algorithm=sampling_algorithm,
            trial=command_job,
            search_space={"ss": Choice(type="choice", values=[{"space1": True}, {"space2": True}])},
            inputs={"input1": Input(path="trial.csv")},
            compute="top_level",
            limits=SweepJobLimits(trial_timeout=600),
        )

        rest = sweep._to_rest_object()
        sweep: SweepJob = Job._from_rest_object(rest)
        assert sweep.sampling_algorithm.type == expected_from_rest_type

    @pytest.mark.parametrize(
        "properties_dict",
        [{}, {"seed": 999}, {"rule": "sobol"}, {"logbase": "e"}, {"seed": 999, "rule": "sobol", "logbase": "e"}],
    )
    def test_random_sampling_object_with_props(self, properties_dict):
        command_job = CommandJob(
            code=Code(base_path="./src"),
            command="python train.py --ss {search_space.ss}",
            inputs={"input1": Input(path="trial.csv")},
            outputs={"default": Output(path="./foo")},
            compute="trial",
            environment="AzureML-sklearn-1.0-ubuntu20.04-py38-cpu:33",
            limits=CommandJobLimits(timeout=120),
        )

        expected_seed = properties_dict.get("seed", None)
        expected_rule = properties_dict.get("rule", None)
        expected_logbase = properties_dict.get("logbase", None)

        random_sampling_algorithm = RandomSamplingAlgorithm(**properties_dict)
        sweep = SweepJob(
            sampling_algorithm=random_sampling_algorithm,
            trial=command_job,
            search_space={"ss": Choice(type="choice", values=[{"space1": True}, {"space2": True}])},
            inputs={"input1": Input(path="trial.csv")},
            compute="top_level",
            limits=SweepJobLimits(trial_timeout=600),
        )

        rest = sweep._to_rest_object()
        assert rest.properties.sampling_algorithm.sampling_algorithm_type == "Random"
        assert rest.properties.sampling_algorithm.seed == expected_seed
        assert rest.properties.sampling_algorithm.rule == expected_rule
        assert rest.properties.sampling_algorithm.logbase == expected_logbase

        sweep: SweepJob = Job._from_rest_object(rest)
        assert sweep.sampling_algorithm.type == "random"
        assert sweep.sampling_algorithm.seed == expected_seed
        assert sweep.sampling_algorithm.rule == expected_rule
        assert sweep.sampling_algorithm.logbase == expected_logbase

    def test_sweep_job_builder_serialization(self) -> None:
        inputs = {
            "uri": Input(
                type=AssetTypes.URI_FILE, path="azureml://datastores/workspaceblobstore/paths/python/data.csv"
            ),
            "lr": LogUniform(min_value=0.001, max_value=0.1),
        }

        node = command(
            description="description",
            environment="AzureML-sklearn-1.0-ubuntu20.04-py38-cpu:33",
            inputs=inputs,
            command="echo ${{inputs.uri}} ${{inputs.lr}} ${{inputs.lr2}}",
            display_name="builder-sweep-job-display",
            compute="testCompute",
            experiment_name="mfe-test1-dataset",
            identity=UserIdentityConfiguration(),
            tags={"tag1": "value1"},
            properties={"prop1": "value1"},
            distribution=MpiDistribution(),
            environment_variables={"EVN1": "VAR1"},
            outputs={"best_model": {}},
            instance_count=2,
            instance_type="STANDARD_BLA",
            timeout=300,
            code="./",
        )

        sweep_node = node.sweep(
            sampling_algorithm="random",
            goal="maximize",
            primary_metric="accuracy",
            compute="sweep-compute",
            max_concurrent_trials=10,
            max_total_trials=100,
            timeout=300,
            trial_timeout=60,
            early_termination_policy=TruncationSelectionPolicy(
                evaluation_interval=100, delay_evaluation=200, truncation_percentage=40
            ),
            search_space={"lr2": LogUniform(min_value=1, max_value=100)},
        )

        # Check that the command property in the original node is kept unchanged.
        assert node.command == "echo ${{inputs.uri}} ${{inputs.lr}} ${{inputs.lr2}}"

        # job name is set in job_operations.py
        assert sweep_node.name is None
        sweep_node.name = "builder-sweep-job"

        trial = CommandJob(
            environment="AzureML-sklearn-1.0-ubuntu20.04-py38-cpu:33",
            command="echo ${{inputs.uri}} ${{search_space.lr}} ${{search_space.lr2}}",
            distribution=MpiDistribution(),
            environment_variables={"EVN1": "VAR1"},
            resources=JobResourceConfiguration(instance_count=2, instance_type="STANDARD_BLA"),
            code="./",
        )

        expected_job = SweepJob(
            trial=trial,
            name="builder-sweep-job",
            description="description",
            display_name="builder-sweep-job-display",
            compute="sweep-compute",
            experiment_name="mfe-test1-dataset",
            identity=UserIdentityConfiguration(),
            tags={"tag1": "value1"},
            properties={"prop1": "value1"},
            objective=Objective(goal="maximize", primary_metric="accuracy"),
            limits=SweepJobLimits(max_concurrent_trials=10, max_total_trials=100, timeout=300, trial_timeout=60),
            sampling_algorithm="random",
            early_termination=TruncationSelectionPolicy(
                evaluation_interval=100, delay_evaluation=200, truncation_percentage=40
            ),
            inputs={
                "uri": Input(
                    type=AssetTypes.URI_FILE, path="azureml://datastores/workspaceblobstore/paths/python/data.csv"
                )
            },
            outputs={"best_model": {}},
            search_space={
                "lr": LogUniform(min_value=0.001, max_value=0.1),
                "lr2": LogUniform(min_value=1, max_value=100),
            },
        )

        assert expected_job._to_dict() == sweep_node._to_job()._to_dict()

    @pytest.mark.parametrize(
        "distribution",
        [
            None,
            MpiDistribution(process_count_per_instance=2),
            PyTorchDistribution(process_count_per_instance=4),
            TensorFlowDistribution(parameter_server_count=2, worker_count=10),
        ],
    )
    def test_sweep_job_trial_distribution_to_rest(self, distribution) -> None:
        command_job = CommandJob(
            code="./src",
            command="python train.py --ss {search_space.ss}",
            distribution=distribution,
            inputs={"input1": Input(path="trial.csv")},
            outputs={"default": Output(path="./foo")},
            compute="trial",
            environment="AzureML-sklearn-1.0-ubuntu20.04-py38-cpu:33",
            limits=CommandJobLimits(timeout=120),
        )

        sweep = SweepJob(
            sampling_algorithm="random",
            trial=command_job,
            search_space={"ss": Choice(type="choice", values=[{"space1": True}, {"space2": True}])},
            inputs={"input1": {"path": "top_level.csv", "type": "uri_file", "mode": "ro_mount"}},
            compute="top_level",
            limits=SweepJobLimits(trial_timeout=600),
            identity=UserIdentityConfiguration(),
        )

        rest_obj = sweep._to_rest_object()
        rest_obj.properties.trial.distribution == distribution._to_rest_object() if distribution else None

        # validate from rest scenario
        sweep_job: SweepJob = Job._from_rest_object(rest_obj)
        assert sweep_job.trial.distribution == sweep.trial.distribution
        assert sweep_job.compute == sweep.compute
        assert sweep_job.limits == sweep.limits
        assert sweep_job.inputs == sweep.inputs
        assert sweep_job.outputs == sweep.outputs
        assert sweep_job.identity == sweep.identity

    @pytest.mark.parametrize(
        "resources",
        [
            (JobResourceConfiguration(instance_type="d2", instance_count=2)),
            (JobResourceConfiguration(instance_type="d2")),
            (JobResourceConfiguration(instance_count=2)),
        ],
    )
    def test_sweep_job_resources_to_rest(self, resources) -> None:
        command_job = CommandJob(
            code="./src",
            command="python train.py --ss {search_space.ss}",
            inputs={"input1": Input(path="trial.csv")},
            outputs={"default": Output(path="./foo")},
            compute="trial",
            environment="AzureML-sklearn-1.0-ubuntu20.04-py38-cpu:33",
            limits=CommandJobLimits(timeout=120),
        )

        sweep = SweepJob(
            sampling_algorithm="random",
            trial=command_job,
            search_space={"ss": Choice(type="choice", values=[{"space1": True}, {"space2": True}])},
            inputs={"input1": {"path": "top_level.csv", "type": "uri_file", "mode": "ro_mount"}},
            compute="top_level",
            limits=SweepJobLimits(trial_timeout=600),
            identity=UserIdentityConfiguration(),
            resources=resources,
        )

        rest_obj = sweep._to_rest_object()
        rest_obj.properties.resources == resources._to_rest_object() if resources else None

        # validate from rest scenario
        sweep_job: SweepJob = Job._from_rest_object(rest_obj)
        assert sweep_job.compute == sweep.compute
        assert sweep_job.limits == sweep.limits
        assert sweep_job.inputs == sweep.inputs
        assert sweep_job.outputs == sweep.outputs
        assert sweep_job.identity == sweep.identity
        if sweep_job.resources:
            if "instance_type" in sweep.resources:
                assert sweep_job.resources.instance_type == sweep.resources["instance_type"]
            if "instance_count" in sweep.resources:
                assert sweep_job.resources.instance_count == sweep.resources["instance_count"]
